########################################################################
# File name: test_network.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import collections
import concurrent.futures
import random
import unittest
import unittest.mock

import dns
import dns.flags

import aioxmpp.network as network

from aioxmpp.testutils import (
    run_coroutine,
    CoroutineMock
)


class Testthreadlocal_resolver_instance(unittest.TestCase):
    def tearDown(self):
        network.reconfigure_resolver()

    def test_get_resolver_returns_Resolver_instance(self):
        self.assertIsInstance(
            network.get_resolver(),
            dns.resolver.Resolver,
        )

    def test_get_resolver_returns_consistent_Resolver_instance(self):
        i1 = network.get_resolver()
        i2 = network.get_resolver()

        self.assertIs(i1, i2)

    def test_get_resolver_is_not_dnspython_default_resolver(self):
        self.assertIsNot(
            network.get_resolver(),
            dns.resolver.get_default_resolver(),
        )

    def test_get_resolver_is_thread_local(self):
        with concurrent.futures.ThreadPoolExecutor(max_workers=1) as executor:
            fut = executor.submit(network.get_resolver)
            done, waiting = concurrent.futures.wait([fut])
            self.assertSetEqual(done, {fut})
            i1 = fut.result()
        i2 = network.get_resolver()
        self.assertIsNot(i1, i2)

    def test_reconfigure_resolver(self):
        with unittest.mock.patch("dns.resolver.Resolver") as Resolver:
            network.reconfigure_resolver()

        Resolver.assert_called_with()

        self.assertEqual(network.get_resolver(), Resolver())

    def test_set_resolver(self):
        network.set_resolver(unittest.mock.sentinel.resolver)
        self.assertIs(
            network.get_resolver(),
            unittest.mock.sentinel.resolver,
        )

    def test_reconfigure_resolver_works_after_set_resolver(self):
        network.set_resolver(unittest.mock.sentinel.resolver)
        with unittest.mock.patch("dns.resolver.Resolver") as Resolver:
            network.reconfigure_resolver()

        Resolver.assert_called_with()

        self.assertEqual(network.get_resolver(), Resolver())


MockSRVRecord = collections.namedtuple(
    "MockRecord",
    [
        "priority",
        "weight",
        "target",
        "port"
    ])

MockTLSARecord = collections.namedtuple(
    "MockRecord",
    [
        "usage",
        "selector",
        "mtype",
        "cert"
    ])

_MockMessage = collections.namedtuple(
    "MockMessage",
    [
        "flags",
    ])


class MockMessage(_MockMessage):
    def __new__(cls, flags=0):
        return _MockMessage.__new__(cls, flags)


class MockAnswer:
    def __init__(self, records, **kwargs):
        self.records = records
        self.response = MockMessage(**kwargs)

    def __iter__(self):
        return iter(self.records)


class MockResolver:
    def __init__(self, tester, actions=[]):
        self.actions = list(actions)
        self.tester = tester
        self.nameservers = ["10.0.0.1"]
        self._flags = None

    def _get_key(self, qname, rdtype, rdclass, tcp):
        result = (qname, rdtype, rdclass)
        if self._strict_tcp:
            result = result + (bool(tcp),)
        if self._flags is not None:
            result += (self._flags,)
        return result

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, tb):
        if exc_type is not None:
            return
        self.finalize()

    def set_flags(self, flags):
        self._flags = flags

    def define_actions(self, action_sequence, strict_tcp=True):
        self._strict_tcp = strict_tcp
        self.actions[:] = action_sequence

    def query(self, qname, rdtype=1, rdclass=1, tcp=False,
              raise_on_no_answer=True, **kwargs):
        if kwargs:
            raise TypeError("Invalid arguments to mock resolver: {}".format(
                ", ".join(kwargs.keys())))

        self.tester.assertTrue(
            self.actions,
            "Unexpected client action (no actions left)")

        key = self._get_key(qname, rdtype, rdclass, tcp)

        next_key, response = self.actions.pop(0)
        self.tester.assertEqual(
            next_key,
            key,
            "Client action mismatch")
        if isinstance(response, Exception):
            raise response

        if raise_on_no_answer and not response:
            raise dns.resolver.NoAnswer()

        return response

    def finalize(self):
        self.tester.assertSequenceEqual(
            self.actions,
            [],
            "client did not execute all actions",
        )


class Testrepeated_query(unittest.TestCase):
    def setUp(self):
        self.tlr = MockResolver(self)
        base = unittest.mock.Mock()
        self.base = base
        self.run_in_executor = unittest.mock.Mock()

        @asyncio.coroutine
        def run_in_executor(executor, func, *args):
            self.run_in_executor(executor, func, *args)
            return func(*args)

        self.patches = [
            unittest.mock.patch(
                "aioxmpp.network.get_resolver",
                new=base.get_resolver,
            ),
            unittest.mock.patch(
                "aioxmpp.network.reconfigure_resolver",
                new=base.reconfigure_resolver,
            ),
            unittest.mock.patch.object(
                asyncio.get_event_loop(),
                "run_in_executor",
                new=run_in_executor
            )
        ]

        self.answer = MockAnswer([])

        # ensure consistent state
        network.reconfigure_resolver()

        for patch in self.patches:
            patch.start()

        base.get_resolver.return_value = self.tlr

    def tearDown(self):
        for patch in self.patches:
            patch.stop()

        # ensure consistent state
        network.reconfigure_resolver()

    def test_reject_non_positive_number_of_attempts(self):
        with self.assertRaisesRegex(
                ValueError,
                "query cannot succeed with non-positive amount of attempts"):
            run_coroutine(network.repeated_query(
                unittest.mock.sentinel.name,
                unittest.mock.sentinel.rdtype,
                nattempts=0))

    def test_use_resolver_from_arguments(self):
        resolver = MockResolver(self)
        resolver.define_actions([
            (
                (
                    "xn--4ca0bs.example.com",
                    dns.rdatatype.A,
                    dns.rdataclass.IN,
                    False,
                    (dns.flags.RD | dns.flags.AD),
                ),
                self.answer
            )
        ])

        with resolver:
            result = run_coroutine(network.repeated_query(
                "äöü.example.com".encode("idna"),
                dns.rdatatype.A,
                resolver=resolver,
            ))

        self.assertIs(
            result,
            self.answer,
        )

        self.assertSequenceEqual(self.base.mock_calls, [])

    def test_run_query_in_executor(self):
        resolver = MockResolver(self)
        resolver.define_actions([
            (
                (
                    "xn--4ca0bs.example.com",
                    dns.rdatatype.A,
                    dns.rdataclass.IN,
                    False,
                    (dns.flags.RD | dns.flags.AD),
                ),
                self.answer
            )
        ])

        with resolver:
            result = run_coroutine(network.repeated_query(
                "äöü.example.com".encode("idna"),
                dns.rdatatype.A,
                resolver=resolver,
                executor=unittest.mock.sentinel.executor
            ))

        self.assertIs(
            result,
            self.answer,
        )

        self.assertIn(
            unittest.mock.call(
                unittest.mock.sentinel.executor,
                unittest.mock.ANY,
            ),
            self.run_in_executor.mock_calls,
        )

    def test_run_query_in_default_executor_by_default(self):
        resolver = MockResolver(self)
        resolver.define_actions([
            (
                (
                    "xn--4ca0bs.example.com",
                    dns.rdatatype.A,
                    dns.rdataclass.IN,
                    False,
                    (dns.flags.RD | dns.flags.AD),
                ),
                self.answer
            )
        ])

        with resolver:
            result = run_coroutine(network.repeated_query(
                "äöü.example.com".encode("idna"),
                dns.rdatatype.A,
                resolver=resolver,
            ))

        self.assertIs(
            result,
            self.answer,
        )

        self.run_in_executor.assert_called_with(
            None,
            unittest.mock.ANY,
        )

    def test_retry_with_tcp_on_first_timeout_with_fixed_resolver(self):
        resolver = MockResolver(self)
        resolver.define_actions([
            (
                (
                    "xn--4ca0bs.example.com",
                    dns.rdatatype.A,
                    dns.rdataclass.IN,
                    False,
                    (dns.flags.RD | dns.flags.AD),
                ),
                dns.resolver.Timeout()
            ),
            (
                (
                    "xn--4ca0bs.example.com",
                    dns.rdatatype.A,
                    dns.rdataclass.IN,
                    True,
                    (dns.flags.RD | dns.flags.AD),
                ),
                self.answer,
            )
        ])

        with resolver:
            result = run_coroutine(network.repeated_query(
                "äöü.example.com".encode("idna"),
                dns.rdatatype.A,
                resolver=resolver,
            ))

        self.assertIs(
            result,
            self.answer,
        )

        self.assertSequenceEqual(self.base.mock_calls, [])

    def test_retry_up_to_2_times_with_fixed_resolver(self):
        resolver = MockResolver(self)
        resolver.define_actions([
            (
                (
                    "xn--4ca0bs.example.com",
                    dns.rdatatype.A,
                    dns.rdataclass.IN,
                    False,
                    (dns.flags.RD | dns.flags.AD),
                ),
                dns.resolver.Timeout()
            ),
            (
                (
                    "xn--4ca0bs.example.com",
                    dns.rdatatype.A,
                    dns.rdataclass.IN,
                    True,
                    (dns.flags.RD | dns.flags.AD),
                ),
                dns.resolver.Timeout()
            ),
        ])

        with resolver:
            with self.assertRaises(TimeoutError):
                run_coroutine(network.repeated_query(
                    "äöü.example.com".encode("idna"),
                    dns.rdatatype.A,
                    resolver=resolver,
                ))

        self.assertSequenceEqual(self.base.mock_calls, [])

    def test_use_thread_local_resolver(self):
        self.tlr.define_actions([
            (
                (
                    "xn--4ca0bs.example.com",
                    dns.rdatatype.A,
                    dns.rdataclass.IN,
                    False,
                    (dns.flags.RD | dns.flags.AD),
                ),
                self.answer,
            )
        ])

        with self.tlr:
            result = run_coroutine(network.repeated_query(
                "äöü.example.com".encode("idna"),
                dns.rdatatype.A,
            ))

        self.assertSequenceEqual(
            self.base.mock_calls,
            [
                unittest.mock.call.get_resolver(),
            ]
        )

        self.assertIs(
            result,
            self.answer,
        )

    def test_reconfigure_resolver_after_first_timeout(self):
        def reconfigure():
            self.tlr.set_flags(None)
            self.tlr.define_actions([
                (
                    (
                        "xn--4ca0bs.example.com",
                        dns.rdatatype.A,
                        dns.rdataclass.IN,
                        False,
                        (dns.flags.RD | dns.flags.AD),
                    ),
                    self.answer,
                )
            ])

        self.base.reconfigure_resolver.side_effect = reconfigure

        self.tlr.define_actions([
            (
                (
                    "xn--4ca0bs.example.com",
                    dns.rdatatype.A,
                    dns.rdataclass.IN,
                    False,
                    (dns.flags.RD | dns.flags.AD),
                ),
                dns.resolver.Timeout(),
            )
        ])

        with self.tlr:
            result = run_coroutine(network.repeated_query(
                "äöü.example.com".encode("idna"),
                dns.rdatatype.A,
            ))

        self.assertSequenceEqual(
            self.base.mock_calls,
            [
                unittest.mock.call.get_resolver(),
                unittest.mock.call.reconfigure_resolver(),
                unittest.mock.call.get_resolver(),
            ]
        )

        self.assertIs(
            result,
            self.answer,
        )

    def test_reconfigure_resolver_after_first_NoNameservers(self):
        def reconfigure():
            self.tlr.set_flags(None)
            self.tlr.define_actions([
                (
                    (
                        "xn--4ca0bs.example.com",
                        dns.rdatatype.A,
                        dns.rdataclass.IN,
                        False,
                        (dns.flags.RD | dns.flags.AD),
                    ),
                    self.answer,
                )
            ])

        self.base.reconfigure_resolver.side_effect = reconfigure

        self.tlr.define_actions([
            (
                (
                    "xn--4ca0bs.example.com",
                    dns.rdatatype.A,
                    dns.rdataclass.IN,
                    False,
                    (dns.flags.RD | dns.flags.AD),
                ),
                dns.resolver.NoNameservers(),
            )
        ])

        with self.tlr:
            result = run_coroutine(network.repeated_query(
                "äöü.example.com".encode("idna"),
                dns.rdatatype.A,
            ))

        self.assertSequenceEqual(
            self.base.mock_calls,
            [
                unittest.mock.call.get_resolver(),
                unittest.mock.call.reconfigure_resolver(),
                unittest.mock.call.get_resolver(),
            ]
        )

        self.assertIs(
            result,
            self.answer,
        )

    def test_use_tcp_after_second_timeout(self):
        def reconfigure():
            self.tlr.set_flags(None)
            self.tlr.define_actions([
                (
                    (
                        "xn--4ca0bs.example.com",
                        dns.rdatatype.A,
                        dns.rdataclass.IN,
                        False,
                        (dns.flags.RD | dns.flags.AD),
                    ),
                    dns.resolver.Timeout(),
                ),
                (
                    (
                        "xn--4ca0bs.example.com",
                        dns.rdatatype.A,
                        dns.rdataclass.IN,
                        True,
                        (dns.flags.RD | dns.flags.AD),
                    ),
                    self.answer,
                )
            ])

        self.base.reconfigure_resolver.side_effect = reconfigure

        self.tlr.define_actions([
            (
                (
                    "xn--4ca0bs.example.com",
                    dns.rdatatype.A,
                    dns.rdataclass.IN,
                    False,
                    (dns.flags.RD | dns.flags.AD),
                ),
                dns.resolver.Timeout(),
            )
        ])

        with self.tlr:
            result = run_coroutine(network.repeated_query(
                "äöü.example.com".encode("idna"),
                dns.rdatatype.A,
            ))

        self.assertSequenceEqual(
            self.base.mock_calls,
            [
                unittest.mock.call.get_resolver(),
                unittest.mock.call.reconfigure_resolver(),
                unittest.mock.call.get_resolver(),
            ]
        )

        self.assertIs(
            result,
            self.answer,
        )

    def test_retry_up_to_3_times_with_thread_local_resolver(self):
        def reconfigure():
            self.tlr.set_flags(None)
            self.tlr.define_actions([
                (
                    (
                        "xn--4ca0bs.example.com",
                        dns.rdatatype.A,
                        dns.rdataclass.IN,
                        False,
                        (dns.flags.RD | dns.flags.AD),
                    ),
                    dns.resolver.Timeout(),
                ),
                (
                    (
                        "xn--4ca0bs.example.com",
                        dns.rdatatype.A,
                        dns.rdataclass.IN,
                        True,
                        (dns.flags.RD | dns.flags.AD),
                    ),
                    dns.resolver.Timeout(),
                )
            ])

        self.base.reconfigure_resolver.side_effect = reconfigure

        self.tlr.define_actions([
            (
                (
                    "xn--4ca0bs.example.com",
                    dns.rdatatype.A,
                    dns.rdataclass.IN,
                    False,
                    (dns.flags.RD | dns.flags.AD),
                ),
                dns.resolver.Timeout(),
            )
        ])

        with self.tlr:
            with self.assertRaises(TimeoutError):
                run_coroutine(network.repeated_query(
                    "äöü.example.com".encode("idna"),
                    dns.rdatatype.A,
                ))

        self.assertSequenceEqual(
            self.base.mock_calls,
            [
                unittest.mock.call.get_resolver(),
                unittest.mock.call.reconfigure_resolver(),
                unittest.mock.call.get_resolver(),
            ]
        )

    def test_overridden_thread_local_behaves_like_fixed(self):
        resolver = MockResolver(self)
        resolver.define_actions([
            (
                (
                    "xn--4ca0bs.example.com",
                    dns.rdatatype.A,
                    dns.rdataclass.IN,
                    False,
                    (dns.flags.RD | dns.flags.AD),
                ),
                dns.resolver.Timeout()
            ),
            (
                (
                    "xn--4ca0bs.example.com",
                    dns.rdatatype.A,
                    dns.rdataclass.IN,
                    True,
                    (dns.flags.RD | dns.flags.AD),
                ),
                dns.resolver.Timeout()
            ),
        ])

        network.set_resolver(resolver)
        self.base.get_resolver.return_value = resolver

        with resolver:
            with self.assertRaises(TimeoutError):
                run_coroutine(network.repeated_query(
                    "äöü.example.com".encode("idna"),
                    dns.rdatatype.A,
                ))

        self.assertSequenceEqual(
            self.base.mock_calls,
            [
                unittest.mock.call.get_resolver(),
            ]
        )

    def test_raise_ValueError_if_AD_not_present_with_require_ad(self):
        self.tlr.define_actions([
            (
                (
                    "xn--4ca0bs.example.com",
                    dns.rdatatype.A,
                    dns.rdataclass.IN,
                    False,
                    (dns.flags.RD | dns.flags.AD),
                ),
                self.answer,
            )
        ])

        with self.tlr:
            with self.assertRaisesRegex(
                    ValueError,
                    "DNSSEC validation not available"):
                run_coroutine(network.repeated_query(
                    "äöü.example.com".encode("idna"),
                    dns.rdatatype.A,
                    require_ad=True,
                ))

    def test_pass_if_AD_present_with_require_ad(self):
        answer = MockAnswer(
            [],
            flags=dns.flags.AD,
        )

        self.tlr.define_actions([
            (
                (
                    "xn--4ca0bs.example.com",
                    dns.rdatatype.A,
                    dns.rdataclass.IN,
                    False,
                    (dns.flags.RD | dns.flags.AD),
                ),
                answer,
            )
        ])

        with self.tlr:
            result = run_coroutine(network.repeated_query(
                "äöü.example.com".encode("idna"),
                dns.rdatatype.A,
                require_ad=True,
            ))

        self.assertIs(result, answer)

    def test_return_None_on_no_anwser(self):
        self.tlr.define_actions([
            (
                (
                    "xn--4ca0bs.example.com",
                    dns.rdatatype.A,
                    dns.rdataclass.IN,
                    False,
                    (dns.flags.RD | dns.flags.AD),
                ),
                dns.resolver.NoAnswer(),
            )
        ])

        with self.tlr:
            result = run_coroutine(network.repeated_query(
                "äöü.example.com".encode("idna"),
                dns.rdatatype.A,
                require_ad=True,
            ))

        self.assertIsNone(result)

    def test_return_None_on_NXDOMAIN(self):
        self.tlr.define_actions([
            (
                (
                    "xn--4ca0bs.example.com",
                    dns.rdatatype.A,
                    dns.rdataclass.IN,
                    False,
                    (dns.flags.RD | dns.flags.AD),
                ),
                dns.resolver.NXDOMAIN(),
            )
        ])

        with self.tlr:
            result = run_coroutine(network.repeated_query(
                "äöü.example.com".encode("idna"),
                dns.rdatatype.A,
                require_ad=True,
            ))

        self.assertIsNone(result)

    def test_check_with_CD_set_after_NoNameservers(self):
        self.tlr.define_actions([
            (
                (
                    "xn--4ca0bs.example.com",
                    dns.rdatatype.A,
                    dns.rdataclass.IN,
                    False,
                    (dns.flags.RD | dns.flags.AD),
                ),
                dns.resolver.NoNameservers(),
            ),
            # need NoNameservers once more since reconfig will happen inbetween
            (
                (
                    "xn--4ca0bs.example.com",
                    dns.rdatatype.A,
                    dns.rdataclass.IN,
                    False,
                    (dns.flags.RD | dns.flags.AD),
                ),
                dns.resolver.NoNameservers(),
            ),
            (
                (
                    "xn--4ca0bs.example.com",
                    dns.rdatatype.A,
                    dns.rdataclass.IN,
                    False,
                    (dns.flags.RD | dns.flags.AD | dns.flags.CD),
                ),
                self.answer,
            ),
        ])

        with self.tlr:
            with self.assertRaisesRegex(
                    network.ValidationError,
                    "nameserver error, most likely DNSSEC validation failed"):
                run_coroutine(network.repeated_query(
                    "äöü.example.com".encode("idna"),
                    dns.rdatatype.A,
                    require_ad=True,
                ))

    def test_treat_NoAnswer_as_succeeded_query_in_validation_query(self):
        self.tlr.define_actions([
            (
                (
                    "xn--4ca0bs.example.com",
                    dns.rdatatype.A,
                    dns.rdataclass.IN,
                    False,
                    (dns.flags.RD | dns.flags.AD),
                ),
                dns.resolver.NoNameservers(),
            ),
            # need NoNameservers once more since reconfig will happen inbetween
            (
                (
                    "xn--4ca0bs.example.com",
                    dns.rdatatype.A,
                    dns.rdataclass.IN,
                    False,
                    (dns.flags.RD | dns.flags.AD),
                ),
                dns.resolver.NoNameservers(),
            ),
            (
                (
                    "xn--4ca0bs.example.com",
                    dns.rdatatype.A,
                    dns.rdataclass.IN,
                    False,
                    (dns.flags.RD | dns.flags.AD | dns.flags.CD),
                ),
                dns.resolver.NoAnswer(),
            ),
        ])

        with self.tlr:
            with self.assertRaisesRegex(
                    network.ValidationError,
                    "nameserver error, most likely DNSSEC validation failed"):
                run_coroutine(network.repeated_query(
                    "äöü.example.com".encode("idna"),
                    dns.rdatatype.A,
                    require_ad=True,
                ))

    def test_treat_NXDOMAIN_as_succeeded_query_in_validation_query(self):
        self.tlr.define_actions([
            (
                (
                    "xn--4ca0bs.example.com",
                    dns.rdatatype.A,
                    dns.rdataclass.IN,
                    False,
                    (dns.flags.RD | dns.flags.AD),
                ),
                dns.resolver.NoNameservers(),
            ),
            # need NoNameservers once more since reconfig will happen inbetween
            (
                (
                    "xn--4ca0bs.example.com",
                    dns.rdatatype.A,
                    dns.rdataclass.IN,
                    False,
                    (dns.flags.RD | dns.flags.AD),
                ),
                dns.resolver.NoNameservers(),
            ),
            (
                (
                    "xn--4ca0bs.example.com",
                    dns.rdatatype.A,
                    dns.rdataclass.IN,
                    False,
                    (dns.flags.RD | dns.flags.AD | dns.flags.CD),
                ),
                dns.resolver.NXDOMAIN(),
            ),
        ])

        with self.tlr:
            with self.assertRaisesRegex(
                    network.ValidationError,
                    "nameserver error, most likely DNSSEC validation failed"):
                run_coroutine(network.repeated_query(
                    "äöü.example.com".encode("idna"),
                    dns.rdatatype.A,
                    require_ad=True,
                ))

    def test_continue_as_normal_on_timeout_after_NoNameservers(self):
        def reconfigure():
            self.tlr.set_flags(None)
            self.tlr.define_actions([
                (
                    (
                        "xn--4ca0bs.example.com",
                        dns.rdatatype.A,
                        dns.rdataclass.IN,
                        False,
                        (dns.flags.RD | dns.flags.AD),
                    ),
                    dns.resolver.Timeout(),
                ),
                (
                    (
                        "xn--4ca0bs.example.com",
                        dns.rdatatype.A,
                        dns.rdataclass.IN,
                        True,
                        (dns.flags.RD | dns.flags.AD),
                    ),
                    self.answer
                ),
            ])

        self.base.reconfigure_resolver.side_effect = reconfigure

        self.tlr.define_actions([
            (
                (
                    "xn--4ca0bs.example.com",
                    dns.rdatatype.A,
                    dns.rdataclass.IN,
                    False,
                    (dns.flags.RD | dns.flags.AD),
                ),
                dns.resolver.NoNameservers(),
            ),
            (
                (
                    "xn--4ca0bs.example.com",
                    dns.rdatatype.A,
                    dns.rdataclass.IN,
                    False,
                    (dns.flags.RD | dns.flags.AD | dns.flags.CD),
                ),
                dns.resolver.Timeout(),
            ),
        ])

        with self.tlr:
            result = run_coroutine(network.repeated_query(
                "äöü.example.com".encode("idna"),
                dns.rdatatype.A,
            ))

        self.assertIs(result, self.answer)

    def test_re_raise_NoNameservers_on_validation_query(self):
        self.tlr.define_actions([
            (
                (
                    "xn--4ca0bs.example.com",
                    dns.rdatatype.A,
                    dns.rdataclass.IN,
                    False,
                    (dns.flags.RD | dns.flags.AD),
                ),
                dns.resolver.NoNameservers(),
            ),
            # need NoNameservers once more since reconfig will happen inbetween
            (
                (
                    "xn--4ca0bs.example.com",
                    dns.rdatatype.A,
                    dns.rdataclass.IN,
                    False,
                    (dns.flags.RD | dns.flags.AD),
                ),
                dns.resolver.NoNameservers(),
            ),
            (
                (
                    "xn--4ca0bs.example.com",
                    dns.rdatatype.A,
                    dns.rdataclass.IN,
                    False,
                    (dns.flags.RD | dns.flags.AD | dns.flags.CD),
                ),
                dns.resolver.NoNameservers()
            ),
        ])

        with self.tlr:
            with self.assertRaises(dns.resolver.NoNameservers):
                run_coroutine(network.repeated_query(
                    "äöü.example.com".encode("idna"),
                    dns.rdatatype.A,
                ))


class Testlookup_srv(unittest.TestCase):
    def setUp(self):
        base = unittest.mock.Mock()
        base.repeated_query = CoroutineMock()

        self.base = base
        self.patches = [
            unittest.mock.patch(
                "aioxmpp.network.repeated_query",
                new=base.repeated_query,
            ),
        ]

        for patch in self.patches:
            patch.start()

    def tearDown(self):
        for patch in self.patches:
            patch.stop()

    def test_return_formatted_records(self):
        self.base.repeated_query.return_value = [
            MockSRVRecord(0, 1, "xmpp.foo.test.", 5222),
            MockSRVRecord(2, 1, "xmpp.bar.test.", 5222),
        ]

        self.assertSequenceEqual(
            run_coroutine(network.lookup_srv(
                b"foo.test",
                "xmpp-client",
                executor=unittest.mock.sentinel.executor,
                resolver=unittest.mock.sentinel.resolver,
            )),
            [
                (0, 1, (b"xmpp.foo.test", 5222)),
                (2, 1, (b"xmpp.bar.test", 5222)),
            ]
        )

        self.base.repeated_query.assert_called_with(
            b"_xmpp-client._tcp.foo.test",
            dns.rdatatype.SRV,
            executor=unittest.mock.sentinel.executor,
            resolver=unittest.mock.sentinel.resolver,
        )

    def test_return_None_on_nxdomain(self):
        self.base.repeated_query.return_value = None

        self.assertIsNone(
            run_coroutine(network.lookup_srv(
                b"foo.test",
                "xmpp-client",
            )),
        )

        self.base.repeated_query.assert_called_with(
            b"_xmpp-client._tcp.foo.test",
            dns.rdatatype.SRV,
        )

    def test_raise_ValueError_if_service_not_supported(self):
        self.base.repeated_query.return_value = [
            MockSRVRecord(0, 1, "xmpp.foo.test.", 5222),
            MockSRVRecord(2, 1, ".", 5222),
        ]

        with self.assertRaisesRegex(
                ValueError,
                r"'xmpp-client' over 'tcp' not supported at b'foo\.test'"):
            run_coroutine(network.lookup_srv(
                b"foo.test",
                "xmpp-client",
            ))


class Testlookup_tlsa(unittest.TestCase):
    def setUp(self):
        base = unittest.mock.Mock()
        base.repeated_query = CoroutineMock()

        self.base = base
        self.patches = [
            unittest.mock.patch(
                "aioxmpp.network.repeated_query",
                new=base.repeated_query,
            ),
        ]

        for patch in self.patches:
            patch.start()

    def tearDown(self):
        for patch in self.patches:
            patch.stop()

    def test_return_formatted_records(self):
        self.base.repeated_query.return_value = [
            MockTLSARecord(3, 0, 1, b"foo"),
            MockTLSARecord(3, 2, 1, b"bar"),
        ]

        self.assertSequenceEqual(
            run_coroutine(network.lookup_tlsa(
                b"foo.test",
                5222,
                executor=unittest.mock.sentinel.executor,
                resolver=unittest.mock.sentinel.resolver,
            )),
            [
                (3, 0, 1, b"foo"),
                (3, 2, 1, b"bar"),
            ]
        )

        self.base.repeated_query.assert_called_with(
            b"_5222._tcp.foo.test",
            dns.rdatatype.TLSA,
            require_ad=True,
            executor=unittest.mock.sentinel.executor,
            resolver=unittest.mock.sentinel.resolver,
        )

    def test_return_None_on_nxdomain(self):
        self.base.repeated_query.return_value = None

        self.assertIsNone(
            run_coroutine(network.lookup_tlsa(
                b"foo.test",
                5222,
            )),
        )

        self.base.repeated_query.assert_called_with(
            b"_5222._tcp.foo.test",
            dns.rdatatype.TLSA,
            require_ad=True,
        )


class Testgroup_and_order_srv_records(unittest.TestCase):
    def _test_monte_carlo_ex(self, hosts, records, N=100):
        rng = random.Random()
        rng.seed(1234)

        host_map = {
            host: collections.Counter()
            for host in hosts
        }

        sum_of_weights = sum(weight for _, weight, _ in records)
        for i in range(0, sum_of_weights*N):
            result = network.group_and_order_srv_records(records, rng=rng)

            for i, host in enumerate(result):
                host_map[host][i] += 1

        return {
            host: {
                index: round(indicies[index]/(N*sum_of_weights), 2)
                for index in range(0, len(records))
            }
            for host, indicies in host_map.items()
        }

    def _test_monte_carlo(self, weights, **kwargs):
        hosts = tuple(i for i, _ in enumerate(weights))
        records = [
            (0, weight, host)
            for host, weight in zip(hosts, weights)
        ]
        return self._test_monte_carlo_ex(hosts, records, **kwargs)

    def test_one_record_with_zero_weight(self):
        h1 = (1,)

        records = [
            (0, 0, h1),
        ]

        self.assertSequenceEqual(
            [h1],
            list(network.group_and_order_srv_records(records))
        )

    def test_zero_weight_mixed(self):
        host_chances = self._test_monte_carlo([0, 0, 5])

        # these values are pretty random

        # note however that the first record always has a much higher chance of
        # being picked
        # the second needs TWO consecutive zeros to get elected to the top,
        # which has a chance in the order of 1/W², with W being the total
        # weight, assuming a equally distributed RNG.
        self.assertEqual(
            0.15,
            host_chances[0][0]
        )

        self.assertEqual(
            0.0,
            host_chances[1][0]
        )

        self.assertEqual(
            0.85,
            host_chances[2][0]
        )

    def test_picking_from_two(self):
        host_chances = self._test_monte_carlo([10, 20])

        self.assertEqual(
            0.35,
            host_chances[0][0]
        )
        self.assertEqual(
            0.65,
            host_chances[0][1]
        )

        # the other host is implied in the values above

    def test_group_by_priority(self):
        hosts = tuple(range(3))

        records = [
            (0, 0, hosts[0]),
            (1, 0, hosts[1]),
            (2, 0, hosts[2]),
        ]

        self.assertSequenceEqual(
            hosts,
            list(network.group_and_order_srv_records(records))
        )

    def test_group_by_priority_with_weight(self):
        hosts = tuple(range(4))

        records = [
            (0, 10, hosts[0]),
            (0, 10, hosts[1]),
            (1, 10, hosts[2]),
            (1, 10, hosts[3]),
        ]

        host_chances = self._test_monte_carlo_ex(hosts, records)

        self.assertEqual(
            0.51,
            host_chances[0][0])
        self.assertEqual(
            0.49,
            host_chances[0][1])
        self.assertEqual(
            0.0,
            host_chances[0][2])
        self.assertEqual(
            0.0,
            host_chances[0][3])

        self.assertEqual(
            0.49,
            host_chances[1][0])
        self.assertEqual(
            0.51,
            host_chances[1][1])
        self.assertEqual(
            0.0,
            host_chances[1][2])
        self.assertEqual(
            0.0,
            host_chances[1][3])

        self.assertEqual(
            0.0,
            host_chances[2][0])
        self.assertEqual(
            0.0,
            host_chances[2][1])
        self.assertEqual(
            0.51,
            host_chances[2][2])
        self.assertEqual(
            0.49,
            host_chances[2][3])

        self.assertEqual(
            0.0,
            host_chances[3][0])
        self.assertEqual(
            0.0,
            host_chances[3][1])
        self.assertEqual(
            0.49,
            host_chances[3][2])
        self.assertEqual(
            0.51,
            host_chances[3][3])

    def test_equal_prio_weight_and_uncomparable_object(self):
        o1, o2, = object(), object()

        records = [
            (0, 0, o1),
            (0, 0, o2),
        ]

        self.assertSequenceEqual(
            [o1, o2],
            list(network.group_and_order_srv_records(records))
        )


class Testfind_xmpp_host_addr(unittest.TestCase):
    def test_returns_items_if_available(self):
        base = unittest.mock.Mock()

        nattempts = object()
        items = object()

        with unittest.mock.patch(
                "aioxmpp.network.lookup_srv",
                new=base.lookup_srv) as lookup_srv:
            lookup_srv.return_value = items

            result = run_coroutine(network.find_xmpp_host_addr(
                asyncio.get_event_loop(),
                base.domain,
                attempts=nattempts
            ))

        calls = list(base.mock_calls)
        self.assertSequenceEqual(
            calls,
            [
                unittest.mock.call.domain.encode("IDNA"),
                unittest.mock.call.lookup_srv(
                    service="xmpp-client",
                    domain=base.domain.encode(),
                    nattempts=nattempts
                )
            ]
        )

        self.assertIs(result, items)

    def test_creates_fake_srv_if_no_srvs_available(self):
        base = unittest.mock.Mock()

        nattempts = object()

        with unittest.mock.patch(
                "aioxmpp.network.lookup_srv",
                new=base.lookup_srv) as lookup_srv:
            lookup_srv.return_value = None

            result = run_coroutine(network.find_xmpp_host_addr(
                asyncio.get_event_loop(),
                base.domain,
                attempts=nattempts
            ))

        calls = list(base.mock_calls)
        self.assertSequenceEqual(
            calls,
            [
                unittest.mock.call.domain.encode("IDNA"),
                unittest.mock.call.lookup_srv(
                    service="xmpp-client",
                    domain=base.domain.encode(),
                    nattempts=nattempts
                )
            ]
        )

        self.assertSequenceEqual(
            result,
            [
                (0, 0, (base.domain.encode(), 5222)),
            ]
        )

    def test_propagates_OSError_from_lookup_srv(self):
        base = unittest.mock.Mock()

        nattempts = object()

        with unittest.mock.patch(
                "aioxmpp.network.lookup_srv",
                new=base.lookup_srv) as lookup_srv:
            lookup_srv.side_effect = OSError()

            with self.assertRaises(OSError):
                run_coroutine(network.find_xmpp_host_addr(
                    asyncio.get_event_loop(),
                    base.domain,
                    attempts=nattempts
                ))

    def test_propagates_ValueError_from_lookup_srv(self):
        base = unittest.mock.Mock()

        nattempts = object()

        with unittest.mock.patch(
                "aioxmpp.network.lookup_srv",
                new=base.lookup_srv) as lookup_srv:
            lookup_srv.side_effect = ValueError()

            with self.assertRaises(ValueError):
                run_coroutine(network.find_xmpp_host_addr(
                    asyncio.get_event_loop(),
                    base.domain,
                    attempts=nattempts
                ))



########################################################################
# File name: test_utils.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import unittest
import unittest.mock

from aioxmpp.chatstates import (ChatState, ChatStateManager,  # NOQA
                                DoNotEmit, AlwaysEmit, DiscoverSupport)


class TestStrategies(unittest.TestCase):

    def test_DoNotEmit(self):
        strategy = DoNotEmit()
        self.assertFalse(strategy.sending)
        strategy.reset()
        self.assertFalse(strategy.sending)
        strategy.no_reply()
        self.assertFalse(strategy.sending)

    def test_DiscoverSupport(self):
        strategy = DiscoverSupport()
        self.assertTrue(strategy.sending)
        strategy.no_reply()
        self.assertFalse(strategy.sending)
        strategy.reset()
        self.assertTrue(strategy.sending)

    def test_AlwaysEmit(self):
        strategy = AlwaysEmit()
        self.assertTrue(strategy.sending)
        strategy.reset()
        self.assertTrue(strategy.sending)
        strategy.no_reply()
        self.assertTrue(strategy.sending)


class TestChatStateManager(unittest.TestCase):

    def test_no_reply(self):
        manager = ChatStateManager(unittest.mock.Mock())
        manager.no_reply()
        manager._strategy.no_reply.assert_called_once_with()

    def test_reset(self):
        manager = ChatStateManager(unittest.mock.Mock())
        manager.reset()
        manager._strategy.reset.assert_called_once_with()

    def test_handle(self):
        manager = ChatStateManager(unittest.mock.sentinel)

        self.assertIs(manager.handle(ChatState.ACTIVE), False)
        self.assertIs(manager.handle(ChatState.INACTIVE),
                      unittest.mock.sentinel.sending)
        self.assertIs(manager.handle(ChatState.INACTIVE), False)
        self.assertIs(manager.handle(ChatState.ACTIVE),
                      unittest.mock.sentinel.sending)
        self.assertIs(manager.handle(ChatState.ACTIVE, message=True),
                      unittest.mock.sentinel.sending)

    def test_handle_message_other_than_active_raises(self):
        manager = ChatStateManager(unittest.mock.sentinel)
        for state in ChatState:
            if state != ChatState.ACTIVE:
                with self.assertRaises(ValueError):
                    manager.handle(state, message=True)



########################################################################
# File name: test_i18n.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import contextlib
import unittest
import unittest.mock

from datetime import datetime, timedelta, date

import babel
import babel.dates
import babel.numbers
import pytz
import tzlocal

import aioxmpp.i18n as i18n


class TestLocalizingFormatter(unittest.TestCase):
    def setUp(self):
        self.default_locale = babel.default_locale()
        self.foreign_locale = babel.Locale("it")
        if self.foreign_locale == self.default_locale:
            self.foreign_locale = babel.Locale("de")

        self.local_timezone = tzlocal.get_localzone()
        self.foreign_timezone = pytz.timezone("US/Eastern")
        if self.foreign_timezone == self.local_timezone:
            self.foreign_timezone = pytz.timezone("Europe/Berlin")

    def test_init_default(self):
        formatter = i18n.LocalizingFormatter()
        self.assertEqual(formatter.locale, babel.default_locale())
        self.assertEqual(formatter.tzinfo, self.local_timezone)

    def test_init_with_explicit_locale(self):
        formatter = i18n.LocalizingFormatter(self.foreign_locale)
        self.assertEqual(formatter.locale, self.foreign_locale)

    def test_init_with_explicit_timezone(self):
        formatter = i18n.LocalizingFormatter(self.foreign_locale,
                                             self.foreign_timezone)
        self.assertEqual(formatter.tzinfo, self.foreign_timezone)

    def test_convert_field_datetime_locale(self):
        formatter = i18n.LocalizingFormatter()

        loc = object()

        tzinfo = unittest.mock.Mock()
        dt = datetime.now(tz=pytz.utc)
        with contextlib.ExitStack() as stack:
            format_datetime = stack.enter_context(
                unittest.mock.patch("babel.dates.format_datetime")
            )
            s = formatter.convert_field(dt, "s",
                                        locale=loc,
                                        tzinfo=tzinfo)

        self.assertSequenceEqual(
            [
                unittest.mock.call(tzinfo.normalize(dt), locale=loc),
            ],
            format_datetime.mock_calls
        )

        self.assertEqual(
            format_datetime(),
            s
        )

    def test_convert_field_datetime_default_locale(self):
        formatter = i18n.LocalizingFormatter()

        dt = datetime.now(tz=pytz.utc)
        with unittest.mock.patch(
                "babel.dates.format_datetime") as format_datetime:
            s = formatter.convert_field(dt, "s")

        self.assertSequenceEqual(
            [
                unittest.mock.call(dt, locale=formatter.locale),
            ],
            format_datetime.mock_calls
        )

        self.assertEqual(
            format_datetime(),
            s
        )

    def test_convert_field_datetime_repr(self):
        formatter = i18n.LocalizingFormatter()

        loc = object()

        dt = datetime.now(tz=pytz.utc)
        with unittest.mock.patch(
                "babel.dates.format_datetime") as format_datetime:
            r = formatter.convert_field(dt, "r", locale=loc)

        self.assertEqual(
            repr(dt),
            r
        )
        self.assertSequenceEqual(
            [],
            format_datetime.mock_calls
        )

    def test_convert_field_timedelta_locale(self):
        formatter = i18n.LocalizingFormatter()

        loc = object()

        td = timedelta(seconds=123)
        with unittest.mock.patch(
                "babel.dates.format_timedelta") as format_timedelta:
            s = formatter.convert_field(td, "s", locale=loc)

        self.assertSequenceEqual(
            [
                unittest.mock.call(td, locale=loc),
            ],
            format_timedelta.mock_calls
        )

        self.assertEqual(
            format_timedelta(),
            s
        )

    def test_convert_field_timedelta_default_locale(self):
        formatter = i18n.LocalizingFormatter()

        td = timedelta(seconds=123)
        with unittest.mock.patch(
                "babel.dates.format_timedelta") as format_timedelta:
            s = formatter.convert_field(td, "s")

        self.assertSequenceEqual(
            [
                unittest.mock.call(td, locale=formatter.locale),
            ],
            format_timedelta.mock_calls
        )

        self.assertEqual(
            format_timedelta(),
            s
        )

    def test_convert_field_timedelta_repr(self):
        formatter = i18n.LocalizingFormatter()

        td = timedelta(seconds=123)
        with unittest.mock.patch(
                "babel.dates.format_timedelta") as format_timedelta:
            r = formatter.convert_field(td, "r")

        self.assertSequenceEqual(
            [],
            format_timedelta.mock_calls
        )

        self.assertEqual(
            repr(td),
            r
        )

    def test_convert_field_date_locale(self):
        formatter = i18n.LocalizingFormatter()

        loc = object()

        d = date.today()
        with unittest.mock.patch(
                "babel.dates.format_date") as format_date:
            s = formatter.convert_field(d, "s", locale=loc)

        self.assertSequenceEqual(
            [
                unittest.mock.call(d, locale=loc),
            ],
            format_date.mock_calls
        )

        self.assertEqual(
            format_date(),
            s
        )

    def test_convert_field_date_default_locale(self):
        formatter = i18n.LocalizingFormatter()

        d = date.today()
        with unittest.mock.patch(
                "babel.dates.format_date") as format_date:
            s = formatter.convert_field(d, "s")

        self.assertSequenceEqual(
            [
                unittest.mock.call(d, locale=formatter.locale),
            ],
            format_date.mock_calls
        )

        self.assertEqual(
            format_date(),
            s
        )

    def test_convert_field_date_repr(self):
        formatter = i18n.LocalizingFormatter()

        d = date.today()
        with unittest.mock.patch(
                "babel.dates.format_date") as format_date:
            r = formatter.convert_field(d, "r")

        self.assertSequenceEqual(
            [],
            format_date.mock_calls
        )

        self.assertEqual(
            repr(d),
            r
        )

    def test_convert_field_time_locale(self):
        formatter = i18n.LocalizingFormatter()

        loc = object()

        t = datetime.now(tz=pytz.utc).time()
        with unittest.mock.patch(
                "babel.dates.format_time") as format_time:
            s = formatter.convert_field(t, "s", locale=loc)

        self.assertSequenceEqual(
            [
                unittest.mock.call(t, locale=loc),
            ],
            format_time.mock_calls
        )

        self.assertEqual(
            format_time(),
            s
        )

    def test_convert_field_time_default_locale(self):
        formatter = i18n.LocalizingFormatter()

        t = datetime.now(tz=pytz.utc).timetz()
        with unittest.mock.patch(
                "babel.dates.format_time") as format_time:
            s = formatter.convert_field(t, "s")

        self.assertSequenceEqual(
            [
                unittest.mock.call(t, locale=formatter.locale),
            ],
            format_time.mock_calls
        )

        self.assertEqual(
            format_time(),
            s
        )

    def test_convert_field_time_repr(self):
        formatter = i18n.LocalizingFormatter()

        t = datetime.now(tz=pytz.utc).timetz()
        with unittest.mock.patch(
                "babel.dates.format_time") as format_time:
            r = formatter.convert_field(t, "r")

        self.assertSequenceEqual(
            [],
            format_time.mock_calls
        )

        self.assertEqual(
            repr(t),
            r
        )

    def test_convert_other_types(self):
        formatter = i18n.LocalizingFormatter()

        values = [
            123,
            1.2,
            "foobar",
            b"barbaz",
        ]

        for value in values:
            self.assertEqual(
                str(value),
                formatter.convert_field(value, "s")
            )

        for value in values:
            self.assertEqual(
                repr(value),
                formatter.convert_field(value, "r")
            )

    def test_format_field_datetime_forwards_to_babel(self):
        formatter = i18n.LocalizingFormatter()

        tzinfo = unittest.mock.Mock()
        dt = datetime.now(tz=pytz.utc)
        loc = object()

        with contextlib.ExitStack() as stack:
            format_datetime = stack.enter_context(
                unittest.mock.patch("babel.dates.format_datetime")
            )
            s = formatter.format_field(dt, "full barbaz",
                                       locale=loc,
                                       tzinfo=tzinfo)

        self.assertSequenceEqual(
            [
                unittest.mock.call(tzinfo.normalize(dt),
                                   locale=loc,
                                   format="full barbaz")
            ],
            format_datetime.mock_calls
        )
        self.assertEqual(
            format_datetime(),
            s
        )

    def test_format_field_datetime_defaults_to_babels_default(self):
        formatter = i18n.LocalizingFormatter()

        tzinfo = unittest.mock.Mock()
        dt = datetime.now(tz=pytz.utc)
        loc = object()

        with contextlib.ExitStack() as stack:
            format_datetime = stack.enter_context(
                unittest.mock.patch("babel.dates.format_datetime")
            )
            s = formatter.format_field(dt, "",
                                       locale=loc,
                                       tzinfo=tzinfo)

        self.assertSequenceEqual(
            [
                unittest.mock.call(tzinfo.normalize(dt),
                                   locale=loc)
            ],
            format_datetime.mock_calls
        )
        self.assertEqual(
            format_datetime(),
            s
        )

    def test_format_field_datetime_forwards_to_babel_with_defaults(self):
        formatter = i18n.LocalizingFormatter()

        formatter.tzinfo = unittest.mock.Mock()
        dt = datetime.now(tz=pytz.utc)

        with contextlib.ExitStack() as stack:
            format_datetime = stack.enter_context(
                unittest.mock.patch("babel.dates.format_datetime")
            )
            s = formatter.format_field(dt, "full barbaz")

        self.assertSequenceEqual(
            [
                unittest.mock.call(formatter.tzinfo.normalize(dt),
                                   locale=formatter.locale,
                                   format="full barbaz")
            ],
            format_datetime.mock_calls
        )
        self.assertEqual(
            format_datetime(),
            s
        )

    def test_format_field_datetime_forwards_to_babel_without_tzinfo(self):
        formatter = i18n.LocalizingFormatter()

        formatter.tzinfo = unittest.mock.Mock()
        dt = datetime.utcnow()

        self.assertIsNone(dt.tzinfo)

        with contextlib.ExitStack() as stack:
            format_datetime = stack.enter_context(
                unittest.mock.patch("babel.dates.format_datetime")
            )
            s = formatter.format_field(dt, "full barbaz")

        self.assertSequenceEqual(
            [
                unittest.mock.call(dt,
                                   locale=formatter.locale,
                                   format="full barbaz")
            ],
            format_datetime.mock_calls
        )
        self.assertEqual(
            format_datetime(),
            s
        )

    def test_format_field_timedelta_forwards_to_babel(self):
        formatter = i18n.LocalizingFormatter()

        tzinfo = unittest.mock.Mock()
        td = timedelta(seconds=120)
        loc = object()

        with contextlib.ExitStack() as stack:
            format_timedelta = stack.enter_context(
                unittest.mock.patch("babel.dates.format_timedelta")
            )
            s = formatter.format_field(td, "full barbaz",
                                       locale=loc,
                                       tzinfo=tzinfo)

        self.assertSequenceEqual(
            [
                unittest.mock.call(td,
                                   locale=loc,
                                   format="full barbaz")
            ],
            format_timedelta.mock_calls
        )
        self.assertEqual(
            format_timedelta(),
            s
        )

    def test_format_field_timedelta_defaults_to_babels_default(self):
        formatter = i18n.LocalizingFormatter()

        tzinfo = unittest.mock.Mock()
        td = timedelta(seconds=120)
        loc = object()

        with contextlib.ExitStack() as stack:
            format_timedelta = stack.enter_context(
                unittest.mock.patch("babel.dates.format_timedelta")
            )
            s = formatter.format_field(td, "",
                                       locale=loc,
                                       tzinfo=tzinfo)

        self.assertSequenceEqual(
            [
                unittest.mock.call(td,
                                   locale=loc)
            ],
            format_timedelta.mock_calls
        )
        self.assertEqual(
            format_timedelta(),
            s
        )

    def test_format_field_timedelta_forwards_to_babel_with_defaults(self):
        formatter = i18n.LocalizingFormatter()

        formatter.tzinfo = unittest.mock.Mock()
        td = timedelta(seconds=120)

        with contextlib.ExitStack() as stack:
            format_timedelta = stack.enter_context(
                unittest.mock.patch("babel.dates.format_timedelta")
            )
            s = formatter.format_field(td, "full barbaz")

        self.assertSequenceEqual(
            [
                unittest.mock.call(td,
                                   locale=formatter.locale,
                                   format="full barbaz")
            ],
            format_timedelta.mock_calls
        )
        self.assertEqual(
            format_timedelta(),
            s
        )

    def test_format_field_date_forwards_to_babel(self):
        formatter = i18n.LocalizingFormatter()

        tzinfo = unittest.mock.Mock()
        d = datetime.now(tz=pytz.utc).date()
        loc = object()

        with contextlib.ExitStack() as stack:
            format_date = stack.enter_context(
                unittest.mock.patch("babel.dates.format_date")
            )
            s = formatter.format_field(d, "full barbaz",
                                       locale=loc,
                                       tzinfo=tzinfo)

        self.assertSequenceEqual(
            [
                unittest.mock.call(d,
                                   locale=loc,
                                   format="full barbaz")
            ],
            format_date.mock_calls
        )
        self.assertEqual(
            format_date(),
            s
        )

    def test_format_field_date_defaults_to_babels_default(self):
        formatter = i18n.LocalizingFormatter()

        tzinfo = unittest.mock.Mock()
        d = datetime.now(tz=pytz.utc).date()
        loc = object()

        with contextlib.ExitStack() as stack:
            format_date = stack.enter_context(
                unittest.mock.patch("babel.dates.format_date")
            )
            s = formatter.format_field(d, "",
                                       locale=loc,
                                       tzinfo=tzinfo)

        self.assertSequenceEqual(
            [
                unittest.mock.call(d,
                                   locale=loc)
            ],
            format_date.mock_calls
        )
        self.assertEqual(
            format_date(),
            s
        )

    def test_format_field_date_forwards_to_babel_with_defaults(self):
        formatter = i18n.LocalizingFormatter()

        formatter.tzinfo = unittest.mock.Mock()
        d = datetime.now(tz=pytz.utc).date()

        with contextlib.ExitStack() as stack:
            format_date = stack.enter_context(
                unittest.mock.patch("babel.dates.format_date")
            )
            s = formatter.format_field(d, "full barbaz")

        self.assertSequenceEqual(
            [
                unittest.mock.call(d,
                                   locale=formatter.locale,
                                   format="full barbaz")
            ],
            format_date.mock_calls
        )
        self.assertEqual(
            format_date(),
            s
        )

    def test_format_field_time_forwards_to_babel(self):
        formatter = i18n.LocalizingFormatter()

        tzinfo = unittest.mock.Mock()
        t = datetime.now(tz=pytz.utc).timetz()
        loc = object()

        with contextlib.ExitStack() as stack:
            format_time = stack.enter_context(
                unittest.mock.patch("babel.dates.format_time")
            )
            s = formatter.format_field(t, "full barbaz",
                                       locale=loc,
                                       tzinfo=tzinfo)

        self.assertSequenceEqual(
            [
                unittest.mock.call(t,
                                   locale=loc,
                                   format="full barbaz")
            ],
            format_time.mock_calls
        )
        self.assertEqual(
            format_time(),
            s
        )

    def test_format_field_time_defaults_to_babels_default(self):
        formatter = i18n.LocalizingFormatter()

        tzinfo = unittest.mock.Mock()
        t = datetime.now(tz=pytz.utc).timetz()
        loc = object()

        with contextlib.ExitStack() as stack:
            format_time = stack.enter_context(
                unittest.mock.patch("babel.dates.format_time")
            )
            s = formatter.format_field(t, "",
                                       locale=loc,
                                       tzinfo=tzinfo)

        self.assertSequenceEqual(
            [
                unittest.mock.call(t,
                                   locale=loc)
            ],
            format_time.mock_calls
        )
        self.assertEqual(
            format_time(),
            s
        )

    def test_format_field_time_forwards_to_babel_with_defaults(self):
        formatter = i18n.LocalizingFormatter()

        formatter.tzinfo = unittest.mock.Mock()
        t = datetime.now(tz=pytz.utc).timetz()

        with contextlib.ExitStack() as stack:
            format_time = stack.enter_context(
                unittest.mock.patch("babel.dates.format_time")
            )
            s = formatter.format_field(t, "full barbaz")

        self.assertSequenceEqual(
            [
                unittest.mock.call(t,
                                   locale=formatter.locale,
                                   format="full barbaz")
            ],
            format_time.mock_calls
        )
        self.assertEqual(
            format_time(),
            s
        )

    def test_format_field_time_forwards_to_babel_without_tzinfo(self):
        formatter = i18n.LocalizingFormatter()

        formatter.tzinfo = unittest.mock.Mock()
        t = datetime.now(tz=pytz.utc).time()

        with contextlib.ExitStack() as stack:
            format_time = stack.enter_context(
                unittest.mock.patch("babel.dates.format_time")
            )
            s = formatter.format_field(t, "full barbaz")

        self.assertSequenceEqual(
            [
                unittest.mock.call(t,
                                   locale=formatter.locale,
                                   format="full barbaz")
            ],
            format_time.mock_calls
        )
        self.assertEqual(
            format_time(),
            s
        )

    def test_format_field_number_uses_format_number_on_empty_format(self):
        formatter = i18n.LocalizingFormatter()

        v = 123
        loc = object()

        with contextlib.ExitStack() as stack:
            format_number = stack.enter_context(
                unittest.mock.patch("babel.numbers.format_number")
            )

            s = formatter.format_field(v, "n", locale=loc)

        self.assertSequenceEqual(
            [
                unittest.mock.call(v, locale=loc),
            ],
            format_number.mock_calls
        )
        self.assertEqual(
            format_number(),
            s
        )

    def test_format_field_number_forwards_to_babel_format_decimal(self):
        formatter = i18n.LocalizingFormatter()

        v = 123
        loc = object()

        with contextlib.ExitStack() as stack:
            format_decimal = stack.enter_context(
                unittest.mock.patch("babel.numbers.format_decimal")
            )

            s = formatter.format_field(v, "##.###n", locale=loc)

        self.assertSequenceEqual(
            [
                unittest.mock.call(v, format="##.###", locale=loc),
            ],
            format_decimal.mock_calls
        )
        self.assertEqual(
            format_decimal(),
            s
        )

    def test_format_works_normally(self):
        formatter = i18n.LocalizingFormatter()

        testsets = [
            ("abc {} def {:10d} ghi {:.2f}", ("foo", 12, 34.5), {}),
        ]

        for fmt, args, kwargs in testsets:
            self.assertEqual(
                fmt.format(*args, **kwargs),
                formatter.format(fmt, *args, **kwargs),
            )


class TestLocalizableString(unittest.TestCase):
    def test_init(self):
        s = "foobar baz"
        ls = i18n.LocalizableString(s)
        self.assertNotEqual(ls, s)

        self.assertEqual(ls.singular, s)
        self.assertIsNone(ls.plural)
        self.assertIsNone(ls.number_index)

    def test_init_plural(self):
        sing = "{1:d} thing"
        plural = "{1:d} things"
        ls = i18n.LocalizableString(sing, plural, 1)

        self.assertEqual(
            sing,
            ls.singular,
        )
        self.assertEqual(
            plural,
            ls.plural
        )
        self.assertEqual(
            "1",
            ls.number_index
        )

    def test_init_plural_number_index_defaults_to_zero(self):
        sing = "{:d} thing"
        plural = "{:d} things"
        ls = i18n.LocalizableString(sing, plural)

        self.assertEqual(
            sing,
            ls.singular,
        )
        self.assertEqual(
            plural,
            ls.plural
        )
        self.assertEqual(
            "0",
            ls.number_index
        )

    def test_init_rejects_number_index_without_plural_string(self):
        with self.assertRaisesRegex(ValueError, "plural is required"):
            i18n.LocalizableString("abc", number_index=12)

    def test_compares_to_other_localizable_strings(self):
        s1 = "foobar baz"
        s2 = "fnord"

        ls1 = i18n.LocalizableString(s1)
        ls2 = i18n.LocalizableString(s1)
        ls3 = i18n.LocalizableString(s2)

        self.assertTrue(ls1 == ls2)
        self.assertTrue(ls2 == ls1)
        self.assertFalse(ls1 != ls2)
        self.assertFalse(ls2 != ls1)

        self.assertTrue(ls1 != ls3)
        self.assertTrue(ls3 != ls1)
        self.assertFalse(ls1 == ls3)
        self.assertFalse(ls3 == ls1)

        self.assertTrue(ls2 != ls3)
        self.assertTrue(ls3 != ls2)
        self.assertFalse(ls2 == ls3)
        self.assertFalse(ls3 == ls2)

    def test_compares_to_other_localizable_strings_with_plural(self):
        s1 = "foobar baz"
        s2 = "fnord"

        ls1 = i18n.LocalizableString(s1, "plural")
        ls2 = i18n.LocalizableString(s1)
        ls3 = i18n.LocalizableString(s2, "plural")

        self.assertTrue(ls1 != ls2)
        self.assertTrue(ls2 != ls1)
        self.assertFalse(ls1 == ls2)
        self.assertFalse(ls2 == ls1)

        self.assertTrue(ls1 != ls3)
        self.assertTrue(ls3 != ls1)
        self.assertFalse(ls1 == ls3)
        self.assertFalse(ls3 == ls1)

        self.assertTrue(ls2 != ls3)
        self.assertTrue(ls3 != ls2)
        self.assertFalse(ls2 == ls3)
        self.assertFalse(ls3 == ls2)

    def test_compares_to_other_localizable_strings_with_plural_index(self):
        s1 = "foobar baz"
        s2 = "fnord"

        ls1 = i18n.LocalizableString(s1, "plural")
        ls2 = i18n.LocalizableString(s1, "plural", 1)
        ls3 = i18n.LocalizableString(s2, "plural")

        self.assertTrue(ls1 != ls2)
        self.assertTrue(ls2 != ls1)
        self.assertFalse(ls1 == ls2)
        self.assertFalse(ls2 == ls1)

        self.assertTrue(ls1 != ls3)
        self.assertTrue(ls3 != ls1)
        self.assertFalse(ls1 == ls3)
        self.assertFalse(ls3 == ls1)

        self.assertTrue(ls2 != ls3)
        self.assertTrue(ls3 != ls2)
        self.assertFalse(ls2 == ls3)
        self.assertFalse(ls3 == ls2)

    def test_compares_to_other_localizable_strings_with_plural_eq(self):
        s1 = "foobar baz"
        s2 = "fnord"

        ls1 = i18n.LocalizableString(s1, "plural")
        ls2 = i18n.LocalizableString(s1, "plural")
        ls3 = i18n.LocalizableString(s2, "plural")

        self.assertTrue(ls1 == ls2)
        self.assertTrue(ls2 == ls1)
        self.assertFalse(ls1 != ls2)
        self.assertFalse(ls2 != ls1)

        self.assertTrue(ls1 != ls3)
        self.assertTrue(ls3 != ls1)
        self.assertFalse(ls1 == ls3)
        self.assertFalse(ls3 == ls1)

        self.assertTrue(ls2 != ls3)
        self.assertTrue(ls3 != ls2)
        self.assertFalse(ls2 == ls3)
        self.assertFalse(ls3 == ls2)

    def test_localize(self):
        formatter = i18n.LocalizingFormatter()
        translator = unittest.mock.Mock()
        s = "abc {a} def {} ghi {foo}"
        ls = i18n.LocalizableString(s)

        with unittest.mock.patch.object(formatter, "vformat") as vformat:
            ls.localize(
                formatter,
                translator,
                "foobar",
                a=123,
                foo=34.5
            )

        self.assertSequenceEqual(
            [
                unittest.mock.call.gettext(s)
            ],
            translator.mock_calls
        )
        self.assertSequenceEqual(
            [
                unittest.mock.call(
                    translator.gettext(),
                    ("foobar",),
                    {
                        "a": 123,
                        "foo": 34.5,
                    },
                )
            ],
            vformat.mock_calls
        )

    def test_localize_plural(self):
        formatter = i18n.LocalizingFormatter()
        translator = unittest.mock.Mock()
        s1 = "{a} thing"
        sn = "{a} things"
        ls = i18n.LocalizableString(s1, sn, "a")

        with contextlib.ExitStack() as stack:
            vformat = stack.enter_context(
                unittest.mock.patch.object(formatter, "vformat")
            )
            get_field = stack.enter_context(
                unittest.mock.patch.object(formatter, "get_field")
            )
            get_field.return_value = 123, "a"

            ls.localize(
                formatter,
                translator,
                "foobar",
                a=123,
                foo=34.5
            )

        self.assertSequenceEqual(
            [
                unittest.mock.call("a", ("foobar",), {"a": 123, "foo": 34.5}),
            ],
            get_field.mock_calls
        )
        self.assertSequenceEqual(
            [
                unittest.mock.call.ngettext(s1, sn, 123)
            ],
            translator.mock_calls
        )
        self.assertSequenceEqual(
            [
                unittest.mock.call(
                    translator.ngettext(),
                    ("foobar",),
                    {
                        "a": 123,
                        "foo": 34.5,
                    },
                )
            ],
            vformat.mock_calls
        )

    def test_str_singular(self):
        ls = i18n.LocalizableString("foobar")
        self.assertEqual(
            "foobar",
            str(ls),
        )

    def test_str_plural(self):
        ls = i18n.LocalizableString("foobar", "baz")
        self.assertEqual(
            "foobar",
            str(ls),
        )

    def test_repr_singular(self):
        ls = i18n.LocalizableString("foobar")
        self.assertEqual(
            "_({!r})".format("foobar"),
            repr(ls),
        )

    def test_repr_plural(self):
        ls = i18n.LocalizableString("foobar", "baz", "a")
        self.assertEqual(
            "ngettext({!r}, {!r}, {!r})".format("foobar", "baz", "a"),
            repr(ls),
        )

    def test_immutable(self):
        ls = i18n.LocalizableString("foo", "bar", "baz")
        with self.assertRaises(AttributeError):
            ls.singular = "fnord"
        with self.assertRaises(AttributeError):
            ls.plural = "fnord"
        with self.assertRaises(AttributeError):
            ls.number_index = "fnord"
        with self.assertRaises(AttributeError):
            ls.foobar = "fnord"

    def test_hashable(self):
        s1 = i18n.LocalizableString("foo")
        s2 = i18n.LocalizableString("foo")

        self.assertEqual(hash(s1), hash(s2))


class Test_(unittest.TestCase):
    def test_creates_string(self):
        with unittest.mock.patch(
                "aioxmpp.i18n.LocalizableString"
        ) as LocalizableString:
            s = i18n._("s")

        self.assertSequenceEqual(
            [
                unittest.mock.call("s"),
            ],
            LocalizableString.mock_calls
        )
        self.assertEqual(
            LocalizableString(),
            s
        )


class Testngettext(unittest.TestCase):
    def test_creates_string(self):
        with unittest.mock.patch(
                "aioxmpp.i18n.LocalizableString"
        ) as LocalizableString:
            s = i18n.ngettext("one", "many", "abc")

        self.assertSequenceEqual(
            [
                unittest.mock.call("one", "many", "abc"),
            ],
            LocalizableString.mock_calls
        )
        self.assertEqual(
            LocalizableString(),
            s
        )



########################################################################
# File name: list_adhoc_commands.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio

import aioxmpp.adhoc

from framework import Example, exec_example


class ListAdhocCommands(Example):
    def prepare_argparse(self):
        super().prepare_argparse()

        def jid(s):
            return aioxmpp.JID.fromstr(s)

        self.argparse.add_argument(
            "peer_jid",
            nargs="?",
            metavar="JID",
            default=None,
            help="Entity to ask for ad-hoc commands. Must be a full jid,"
            " defaults to the domain of the local JID (asking the server)"
        )

    def configure(self):
        super().configure()

        self.adhoc_peer_jid = (
            self.args.peer_jid or
            self.g_jid.replace(resource=None, localpart=None)
        )

    @asyncio.coroutine
    def run_simple_example(self):
        adhoc = self.client.summon(aioxmpp.adhoc.AdHocClient)
        for item in (yield from adhoc.get_commands(self.adhoc_peer_jid)):
            print("{}: {}".format(
                item.node,
                item.name
            ))


if __name__ == "__main__":
    exec_example(ListAdhocCommands())



########################################################################
# File name: fields.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import abc
import collections
import copy

import aioxmpp.xso as xso

from . import xso as forms_xso


descriptor_ns = "{jabber:x:data}field"


FIELD_DOCSTRING_TEMPLATE = \
"""
``{field_type.value}`` field ``{var}``

{label}
"""


class BoundField(metaclass=abc.ABCMeta):
    """
    Abstract base class for objects returned by the field descriptors.

    :param field: field descriptor to bind
    :type field: :class:`AbstractField`
    :param instance: form instance to bind to
    :type instance: :class:`object`

    :class:`BoundField` instances represent the connection between the field
    descriptors present at a form *class* and the *instance* of that class.

    They store of course the value of the field for the specific instance, but
    also possible instance-specific overrides for the metadata attributes
    :attr:`desc`, :attr:`label` and :attr:`required` (and possibly
    :attr:`.BoundOptionsField.options`). By default, these attributes return
    the same value as set on the corresponding `field`, but the attributes can
    be set to different values, which only affects the single form instance.

    The use case is to fill these fields with the information obtained from a
    :class:`.Data` XSO when creating a form with :meth:`.Form.from_xso`: it
    allows the fields to behave exactly like the sender specified.

    Deep-copying a :class:`BoundField` deepcopies all attributes, except the
    :attr:`field` and :attr:`instance` attributes. See also :meth:`clone_for`
    for copying a bound field for a new `instance`.

    Subclass overview:

    .. autosummary::

       BoundSingleValueField
       BoundMultiValueField
       BoundSelectField
       BoundMultiSelectField

    Binding relationship:

    .. autoattribute:: field

    .. attribute:: instance

       The `instance` as passed to the constructor.

    Field metadata attributes:

    .. autoattribute:: desc

    .. autoattribute:: label

    .. autoattribute:: required

    Helper methods:

    .. automethod:: clone_for

    The following methods must be implemented by base classes.

    .. automethod:: load

    .. automethod:: render

    """

    def __init__(self, field, instance):
        super().__init__()
        self._field = field
        self.instance = instance

    @property
    def field(self):
        """
        The field which is bound to the :attr:`instance`.
        """
        return self._field

    def __repr__(self):
        return "<bound {!r} for {!r} at 0x{:x}>".format(
            self._field,
            self.instance,
            id(self),
        )

    def __deepcopy__(self, memo):
        result = copy.copy(self)
        for k, v in self.__dict__.items():
            if k == "_field" or k == "instance":
                continue
            setattr(result, k, copy.deepcopy(v, memo))
        return result

    @property
    def desc(self):
        """
        .. seealso::

           :attr:`.Field.desc`
              for a full description of the ``desc`` semantics.
        """

        try:
            return self._desc
        except AttributeError:
            return self._field.desc

    @desc.setter
    def desc(self, value):
        self._desc = value

    @property
    def label(self):
        """
        .. seealso::

           :attr:`.Field.label`
              for a full description of the ``label`` semantics.
        """
        try:
            return self._label
        except AttributeError:
            return self._field.label

    @label.setter
    def label(self, value):
        self._label = value

    @property
    def required(self):
        """
        .. seealso::

           :attr:`.Field.required`
              for a full description of the ``required`` semantics.
        """
        try:
            return self._required
        except AttributeError:
            return self._field.required

    @required.setter
    def required(self, value):
        self._required = value

    def clone_for(self, other_instance, memo=None):
        """
        Clone this bound field for another instance, possibly during a
        :func:`~copy.deepcopy` operation.

        :param other_instance: Another form instance to which the newly created
                               bound field shall be bound.
        :type other_instance: :class:`object`
        :param memo: Optional deepcopy-memo (see :mod:`copy` for details)

        If this is called during a deepcopy operation, passing the `memo` helps
        preserving and preventing loops. This method is essentially a
        deepcopy-operation, with a modification of the :attr:`instance`
        afterwards.
        """

        if memo is None:
            result = copy.deepcopy(self)
        else:
            result = copy.deepcopy(self, memo)
        result.instance = other_instance
        return result

    @abc.abstractmethod
    def load(self, field_xso):
        """
        Load the field information from a data field.

        :param field_xso: XSO describing the field.
        :type field_xso: :class:`~.Field`

        This loads the current value, description, label and possibly options
        from the `field_xso`, shadowing the information from the declaration of
        the field on the class.

        This method is must be overriden and is thus marked abstract. However,
        when called from a subclass, it loads the :attr:`desc`, :attr:`label`
        and :attr:`required` from the given `field_xso`. Subclasses are
        supposed to implement a mechansim to load options and/or values from
        the `field_xso` and then call this implementation through
        :func:`super`.
        """
        if field_xso.desc:
            self._desc = field_xso.desc

        if field_xso.label:
            self._label = field_xso.label

        self._required = field_xso.required

    @abc.abstractmethod
    def render(self, *, use_local_metadata=True):
        """
        Return a :class:`~.Field` containing the values and metadata set in the
        field.

        :param use_local_metadata: if true, the description, label and required
                                   metadata can be sourced from the field
                                   descriptor associated with this bound field.
        :type use_local_metadata: :class:`bool`
        :return: A new :class:`~.Field` instance.

        The returned object uses the values accessible through this object;
        that means, any values set for e.g. :attr:`desc` take precedence over
        the values declared at the class level. If `use_local_metadata` is
        false, values declared at the class level are not used if no local
        values are declared. This is useful when generating a reply to a form
        received by a peer, as it avoids sending a modified form.

        This method is must be overriden and is thus marked abstract. However,
        when called from a subclass, it creates the :class:`~.Field` instance
        and initialises its :attr:`~.Field.var`, :attr:`~.Field.type_`,
        :attr:`~.Field.desc`, :attr:`~.Field.required` and
        :attr:`~.Field.label` attributes and returns the result. Subclasses are
        supposed to override this method, call the base implementation through
        :func:`super` to obtain the :class:`~.Field` instance and then fill in
        the values and/or options.
        """

        result = forms_xso.Field(
            var=self.field.var,
            type_=self.field.FIELD_TYPE,
        )

        if use_local_metadata:
            result.desc = self.desc
            result.label = self.label
            result.required = self.required
        else:
            try:
                result.desc = self._desc
            except AttributeError:
                pass

            try:
                result.label = self._label
            except AttributeError:
                pass

            try:
                result.required = self._required
            except AttributeError:
                pass

        return result


class BoundSingleValueField(BoundField):
    """
    A bound field which has only a single value at any time. Only the first
    value is parsed when loading data from a :class:`~.Field` XSO. When writing
    data to a :class:`~.Field` XSO, :data:`None` is treated as the absence of
    any value; every other value is serialised through the
    :attr:`~.AbstractField.type_` of the field.

    .. seealso::

       :class:`BoundField`
          for a description of the arguments.

    This bound field is used by :class:`TextSingle`, :class:`TextPrivate` and
    :class:`JIDSingle`.

    .. autoattribute:: value
    """

    @property
    def value(self):
        """
        The current value of the field. If no value is set when this attribute
        is accessed for reading, the :meth:`default` of the field is invoked
        and the result is set and returned as value.

        Only values which pass through :meth:`~.AbstractCDataType.coerce` of
        the :attr:`~.AbstractField.type_` of the field can be set. To
        revert the :attr:`value` to its default, use the ``del`` operator.
        """
        try:
            return self._value
        except AttributeError:
            # call through to field
            self._value = self._field.default()
            return self._value

    @value.setter
    def value(self, value):
        self._value = self._field.type_.coerce(value)

    @value.deleter
    def value(self):
        try:
            del self._value
        except AttributeError:
            pass

    def load(self, field_xso):
        try:
            value = field_xso.values[0]
        except IndexError:
            value = self._field.default()
        else:
            value = self._field.type_.parse(value)

        self._value = value

        super().load(field_xso)

    def render(self, **kwargs):
        result = super().render(**kwargs)

        try:
            value = self._value
        except AttributeError:
            value = self._field.default()

        if value is None:
            return result

        result.values[:] = [
            self.field.type_.format(value)
        ]

        return result


class BoundMultiValueField(BoundField):
    """
    A bound field which can have multiple values.

    .. seealso::

       :class:`BoundField`
          for a description of the arguments.

    This bound field is used by :class:`TextMulti` and :class:`JIDMulti`.

    .. autoattribute:: value
    """

    @property
    def value(self):
        """
        A tuple of values. This attribute can be set with any iterable; the
        iterable is then evaluated into a tuple and stored at the bound field.

        Whenever values are written to this attribute, they are passed through
        the :meth:`~.AbstractCDataType.coerce` method of the
        :attr:`~.AbstractField.type_` of the field. To revert the
        :attr:`value` to its default, use the ``del`` operator.
        """
        try:
            return self._value
        except AttributeError:
            self.value = self._field.default()
            return self._value

    @value.setter
    def value(self, values):
        coerce = self._field.type_.coerce
        self._value = tuple(
            coerce(v)
            for v in values
        )

    @value.deleter
    def value(self):
        try:
            del self._value
        except AttributeError:
            pass

    def load(self, field_xso):
        self._value = tuple(
            self._field.type_.parse(v)
            for v in field_xso.values
        )

        super().load(field_xso)

    def render(self, **kwargs):
        result = super().render(**kwargs)
        result.values[:] = (
            self.field.type_.format(v)
            for v in self._value
        )
        return result


class BoundOptionsField(BoundField):
    """
    This is an intermediate base class used to implement bound fields for
    fields which have options from which one or more values must be chosen.

    .. seealso::

       :class:`BoundField`
          for a description of the arguments.

    When the field is loaded from a :class:`~.Field` XSO, the options are also
    loaded from there and thus shadow the options defined at the `field`. This
    may come to a surprise of code expecting a specific set of options.

    Subclass overview:

    .. autosummary::

       BoundSelectField
       BoundMultiSelectField

    .. autoattribute:: options
    """

    @property
    def options(self):
        """
        This is a :class:`collections.OrderedDict` which maps option keys to
        their labels. The keys are used as the values of the field; the labels
        are human-readable text for display.

        This attribute can be written with any object which is compatible with
        the dict-constructor. The order is preserved if a sequence of key-value
        pairs is used.

        When writing the attribute, the keys are checked against the
        :meth:`~.AbstractCDataType.coerce` method of the
        :attr:`~.AbstractField.type_` of the field. To make the :attr:`options`
        attribute identical to the :attr:`~.AbstractField.options` attribute,
        use the ``del`` operator.

        .. warning::

           This attribute is mutable, however, mutating it directly may have
           unexpected side effects:

           * If the attribute has not been set before, you will actually be
             mutating the :class:`~.AbstractChoiceField.options` attributes
             value.

             This may be changed in the future by copying more eagerly.

           * The type checking cannot take place when keys are added by direct
             mutation of the dictionary. This means that errors will be delayed
             until the actual serialisation of the data form, which may be a
             confusing thing to debug.

           Relying on the above behaviour or any other behaviour induced by
           directly mutating the value returned by this attribute is **not
           recommended**. Changes to this behaviour are *not* considered
           breaking changes and will be done without the usual deprecation.
        """
        try:
            return self._options
        except AttributeError:
            return self.field.options

    @options.setter
    def options(self, value):
        iterator = (value.items()
                    if isinstance(value, collections.abc.Mapping)
                    else value)
        self._options = collections.OrderedDict(
            (self.field.type_.coerce(k), v)
            for k, v in iterator
        )

    @options.deleter
    def options(self):
        try:
            del self._options
        except AttributeError:
            pass

    def load(self, field_xso):
        self._options = collections.OrderedDict(
            field_xso.options
        )
        super().load(field_xso)

    def render(self, **kwargs):
        format_ = self._field.type_.format
        field_xso = super().render(**kwargs)
        field_xso.options.update(
            (format_(k), v)
            for k, v in self.options.items()
        )
        return field_xso


class BoundSelectField(BoundOptionsField):
    """
    Bound field carrying one value out of a set of options.

    .. seealso::

       :class:`BoundField`
          for a description of the arguments.
       :attr:`BoundOptionsField.options`
          for semantics and behaviour of the ``options`` attribute

    .. autoattribute:: value

    """

    @property
    def value(self):
        """
        The current value of the field. If no value is set when this attribute
        is accessed for reading, the :meth:`default` of the field is invoked
        and the result is set and returned as value.

        Only values contained in the :attr:`~.BoundOptionsField.options` can be
        set, other values are rejected with a :class:`ValueError`. To revert
        the value to the default value specified in the descriptor, use the
        ``del`` operator.
        """
        try:
            return self._value
        except AttributeError:
            self._value = self.field.default()
            return self._value

    @value.setter
    def value(self, value):
        options = self.options
        if value not in options:
            raise ValueError("{!r} not in field options: {!r}".format(
                value,
                tuple(options.keys()),
            ))

        self._value = value

    @value.deleter
    def value(self):
        try:
            del self._value
        except AttributeError:
            pass

    def load(self, field_xso):
        try:
            value = field_xso.values[0]
        except IndexError:
            try:
                del self._value
            except AttributeError:
                pass
        else:
            self._value = self.field.type_.parse(value)

        super().load(field_xso)

    def render(self, **kwargs):
        format_ = self._field.type_.format
        field_xso = super().render(**kwargs)
        value = self.value
        if value is not None:
            field_xso.values[:] = [format_(value)]
        return field_xso


class BoundMultiSelectField(BoundOptionsField):
    """
    Bound field carrying a subset of values out of a set of options.

    .. seealso::

       :class:`BoundField`
          for a description of the arguments.
       :attr:`BoundOptionsField.options`
          for semantics and behaviour of the ``options`` attribute

    .. autoattribute:: value
    """

    @property
    def value(self):
        """
        A :class:`frozenset` whose elements are a subset of the keys of the
        :attr:`~.BoundOptionsField.options` mapping.

        This value can be written with any iterable; the iterable is then
        evaluated into a :class:`frozenset`. If it contains any value not
        contained in the set of keys of options, the attribute is not written
        and :class:`ValueError` is raised.

        To revert the value to the default specified by the field descriptor,
        use the ``del`` operator.
        """
        try:
            return self._value
        except AttributeError:
            self.value = self.field.default()
            return self._value

    @value.setter
    def value(self, values):
        new_values = frozenset(values)
        options = set(self.options.keys())
        invalid = new_values - options
        if invalid:
            raise ValueError(
                "{!r} not in field options: {!r}".format(
                    next(iter(invalid)),
                    tuple(self.options.keys())
                )
            )

        self._value = new_values

    @value.deleter
    def value(self):
        try:
            del self._value
        except AttributeError:
            pass

    def load(self, field_xso):
        self._value = frozenset(
            self.field.type_.parse(value)
            for value in field_xso.values
        )
        super().load(field_xso)

    def render(self, **kwargs):
        format_ = self.field.type_.format

        result = super().render(**kwargs)
        result.values[:] = [
            format_(value)
            for value in self.value
        ]
        return result


class AbstractDescriptor(metaclass=abc.ABCMeta):
    attribute_name = None
    root_class = None

    @abc.abstractmethod
    def descriptor_keys(self):
        """
        Return an iterator with the descriptor keys for this descriptor. The
        keys will be added to the :attr:`DescriptorClass.DESCRIPTOR_KEYS`
        mapping, pointing to the descriptor.

        Duplicate keys will lead to a :class:`TypeError` being raised during
        declaration of the class.
        """


class AbstractField(AbstractDescriptor):
    """
    Abstract base class to implement field descriptor classes.

    :param var: The field ``var`` attribute this descriptor is supposed to
                represent.
    :type var: :class:`str`
    :param type_: The type of the data, defaults to :class:`~.xso.String`.
    :type type_: :class:`~.xso.AbstractCDataType`
    :param required: Flag to indicate that the field is required.
    :type required: :class:`bool`
    :param desc: Description text for the field, e.g. for tool-tips.
    :type desc: :class:`str`, without newlines
    :param label: Short, human-readable label for the field
    :type label: :class:`str`

    The arguments are used to initialise the respective attributes. Details on
    the semantics can be found in the respective documentation pieces below.

    .. autoattribute:: desc

    .. attribute:: label

       Represents the label flag as specified per :xep:`4`.

       The value of this attribute is used when forms are generated locally.
       When forms are received from remote peers and :class:`~.Form` instances
       are constructed from that data, this attribute is not used when
       rendering a reply or when the value is accessed through the bound field.

        .. seealso::

           :attr:`~.Field.label`
              for details on the semantics of this attribute

    .. attribute:: required

       Represents the required flag as specified per :xep:`4`.

       The value of this attribute is used when forms are generated locally.
       When forms are received from remote peers and :class:`~.Form` instances
       are constructed from that data, this attribute is not used when
       rendering a reply or when the value is accessed through the bound field.

        .. seealso::

           :attr:`~.Field.required`
              for details on the semantics of this attribute

    .. autoattribute:: var

    .. automethod:: create_bound

    .. automethod:: default

    .. automethod:: make_bound
    """

    def __init__(self, var, type_, *, required=False, desc=None, label=None):
        super().__init__()
        self._var = var
        self.required = required
        self.desc = desc
        self.label = label
        self._type = type_
        self.__doc__ = FIELD_DOCSTRING_TEMPLATE.format(
            field_type=self.FIELD_TYPE,
            var=self.var,
            desc=self.desc,
            label=self.label,
        )

    def descriptor_keys(self):
        yield descriptor_ns, self._var

    @property
    def type_(self):
        """
        :class:`.AbstractCDataType` instance used to parse, validate and
        format the value(s) of this field.

        The type of a field cannot be changed after its initialisation.
        """
        return self._type

    @property
    def desc(self):
        """
        Represents the description as specified per :xep:`4`.

        The value of this attribute is used when forms are generated locally.
        When forms are received from remote peers and :class:`~.Form` instances
        are constructed from that data, this attribute is not used when
        rendering a reply or when the value is accessed through the bound
        field.

        .. seealso::

           :attr:`~.Field.desc`
              for details on the semantics of this attribute
        """

        return self._desc

    @desc.setter
    def desc(self, value):
        if value is not None and any(ch == "\r" or ch == "\n" for ch in value):
            raise ValueError("desc must not contain newlines")
        self._desc = value

    @desc.deleter
    def desc(self):
        self._desc = None

    @property
    def var(self):
        """
        Represents the field ID as specified per :xep:`4`.

        The value of this attribute is used to match fields when instantiating
        :class:`~.Form` classes from :class:`~.Data` XSOs.

        .. seealso::

           :attr:`~.Field.var`
              for details on the semantics of this attribute
        """
        return self._var

    @abc.abstractmethod
    def default(self):
        """
        Create and return a default value for this field.

        This must be implemented by subclasses.
        """

    @abc.abstractmethod
    def create_bound(self, for_instance):
        """
        Create a :ref:`bound field class <api-aioxmpp.forms-bound-fields>`
        instance for this field for the given form object and return it.

        :param for_instance: The form instance to which the bound field should
                             be bound.

        This method must be re-implemented by subclasses.

        .. seealso::

           :meth:`make_bound`
              creates (using this method) or returns an existing bound field
              for a given form instance.

        """

    def make_bound(self, for_instance):
        """
        Create a new :ref:`bound field class <api-aioxmpp.forms-bound-fields>`
        or return an existing one for the given form object.

        :param for_instance: The form instance to which the bound field should
                             be bound.

        If no bound field can be found on the given `for_instance` for this
        field, a new one is created using :meth:`create_bound`, stored at the
        instance and returned. Otherwise, the existing instance is returned.

        .. seealso::

           :meth:`create_bound`
              creates a new bound field for the given form instance (without
              storing it anywhere).
        """

        try:
            return for_instance._descriptor_data[self]
        except KeyError:
            bound = self.create_bound(for_instance)
            for_instance._descriptor_data[self] = bound
            return bound

    def __get__(self, instance, type_):
        if instance is None:
            return self
        return self.make_bound(instance)


class TextSingle(AbstractField):
    """
    Represent a ``"text-single"`` input with the given `var`.

    :param default: A default value to initialise the field.

    .. seealso::

       :class:`~.fields.BoundSingleValueField`
          is the :ref:`bound field class <api-aioxmpp.forms-bound-fields>` used
          by fields of this type.

       :class:`~.fields.AbstractField`
          for documentation on the `var`, `type_`, `required`, `desc` and
          `label` arguments.

    """
    FIELD_TYPE = forms_xso.FieldType.TEXT_SINGLE

    def __init__(self, var, type_=xso.String(), *,
                 default=None,
                 **kwargs):
        super().__init__(var, type_, **kwargs)
        self._default = default

    def default(self):
        return self._default

    def create_bound(self, for_instance):
        return BoundSingleValueField(
            self,
            for_instance,
        )


class JIDSingle(AbstractField):
    """
    Represent a ``"jid-single"`` input with the given `var`.

    :param default: A default value to initialise the field.

    .. seealso::

       :class:`~.fields.BoundSingleValueField`
          is the :ref:`bound field class <api-aioxmpp.forms-bound-fields>` used
          by fields of this type.

       :class:`~.fields.AbstractField`
          for documentation on the `var`, `required`, `desc` and
          `label` arguments.

    """
    FIELD_TYPE = forms_xso.FieldType.JID_SINGLE

    def __init__(self, var, *, default=None, **kwargs):
        super().__init__(var, type_=xso.JID(), **kwargs)
        self._default = default

    def default(self):
        return self._default

    def create_bound(self, for_instance):
        return BoundSingleValueField(
            self,
            for_instance,
        )


class Boolean(AbstractField):
    """
    Represent a ``"boolean"`` input with the given `var`.

    :param default: A default value to initialise the field.

    .. seealso::

       :class:`~.fields.BoundSingleValueField`
          is the :ref:`bound field class <api-aioxmpp.forms-bound-fields>` used
          by fields of this type.

       :class:`~.fields.AbstractField`
          for documentation on the `var`, `required`, `desc` and
          `label` arguments.

    """

    FIELD_TYPE = forms_xso.FieldType.BOOLEAN

    def __init__(self, var, *, default=False, **kwargs):
        super().__init__(var, xso.Bool(), **kwargs)
        self._default = default

    def default(self):
        return self._default

    def create_bound(self, for_instance):
        return BoundSingleValueField(
            self,
            for_instance,
        )


class TextPrivate(TextSingle):
    """
    Represent a ``"text-private"`` input with the given `var`.

    :param default: A default value to initialise the field.

    .. seealso::

       :class:`~.fields.BoundSingleValueField`
          is the :ref:`bound field class <api-aioxmpp.forms-bound-fields>` used
          by fields of this type.

       :class:`~.fields.AbstractField`
          for documentation on the `var`, `type_`, `required`, `desc` and
          `label` arguments.

    """

    FIELD_TYPE = forms_xso.FieldType.TEXT_PRIVATE


class TextMulti(AbstractField):
    """
    Represent a ``"text-multi"`` input with the given `var`.

    :param default: A default value to initialise the field.
    :type default: :class:`tuple`

    .. seealso::

       :class:`~.fields.BoundMultiValueField`
          is the :ref:`bound field class <api-aioxmpp.forms-bound-fields>` used
          by fields of this type.

       :class:`~.fields.AbstractField`
          for documentation on the `var`, `type_`, `required`, `desc` and
          `label` arguments.

    """

    FIELD_TYPE = forms_xso.FieldType.TEXT_MULTI

    def __init__(self, var, type_=xso.String(), *,
                 default=(), **kwargs):
        super().__init__(var, type_, **kwargs)
        self._default = default

    def create_bound(self, for_instance):
        return BoundMultiValueField(self, for_instance)

    def default(self):
        return self._default


class JIDMulti(AbstractField):
    """
    Represent a ``"jid-multi"`` input with the given `var`.

    :param default: A default value to initialise the field.
    :type default: :class:`tuple`

    .. seealso::

       :class:`~.fields.BoundMultiValueField`
          is the :ref:`bound field class <api-aioxmpp.forms-bound-fields>` used
          by fields of this type.

       :class:`~.fields.AbstractField`
          for documentation on the `var`, `type_`, `required`, `desc` and
          `label` arguments.

    """

    FIELD_TYPE = forms_xso.FieldType.JID_MULTI

    def __init__(self, var, *, default=(), **kwargs):
        super().__init__(var, xso.JID(), **kwargs)
        self._default = default

    def create_bound(self, for_instance):
        return BoundMultiValueField(self, for_instance)

    def default(self):
        return self._default


class AbstractChoiceField(AbstractField):
    """
    Abstract base class to implement field descriptor classes using options.

    :param type_: Type used for the option keys.
    :type type_: :class:`~.xso.AbstractCDataType`
    :param options: A sequence of key-value pairs or a mapping object
                    representing the options available.
    :type options: sequence of pairs or mapping

    The keys of the `options` mapping (or the first elements in the pairs in
    the sequence of pairs) must be compatible with `type_`, in the sense that
    must pass through :meth:`~.xso.AbstractCDataType.coerce` (this is enforced
    when the field is instantiated).

    Fields using this base class:

    .. autosummary::

       aioxmpp.forms.ListSingle
       aioxmpp.forms.ListMulti

    .. seealso::

       :class:`~.fields.BoundOptionsField`
          is an abstract base class to implement :ref:`bound field classes
          <api-aioxmpp.forms-bound-fields>` for fields inheriting from this
          class.

       :class:`~.fields.AbstractField`
          for documentation on the `var`, `required`, `desc` and `label`
          arguments.

    """

    def __init__(self, var, *,
                 type_=xso.String(),
                 options=[],
                 **kwargs):
        super().__init__(var, type_, **kwargs)
        iterator = (options.items()
                    if isinstance(options, collections.abc.Mapping)
                    else options)
        self.options = collections.OrderedDict(
            (type_.coerce(k), v)
            for k, v in iterator
        )
        self._type = type_

    @property
    def type_(self):
        return self._type


class ListSingle(AbstractChoiceField):
    """
    Represent a ``"list-single"`` input with the given `var`.

    :param default: A default value to initialise the field. This must be a
                    member of the `options`.

    .. seealso::

       :class:`~.fields.BoundMultiValueField`
          is the :ref:`bound field class <api-aioxmpp.forms-bound-fields>` used
          by fields of this type.

       :class:`~.fields.AbstractChoiceField`
          for documentation on the `options` argument.

       :class:`~.fields.AbstractField`
          for documentation on the `var`, `type_`, `required`, `desc` and
          `label` arguments.

    """

    FIELD_TYPE = forms_xso.FieldType.LIST_SINGLE

    def __init__(self, var, *, default=None, **kwargs):
        super().__init__(var, **kwargs)
        if default is not None and default not in self.options:
            raise ValueError("invalid default: not in options")
        self._default = default

    def create_bound(self, for_instance):
        return BoundSelectField(self, for_instance)

    def default(self):
        return self._default


class ListMulti(AbstractChoiceField):
    """
    Represent a ``"list-multi"`` input with the given `var`.

    :param default: An iterable of `options` keys
    :type default: iterable

    `default` is evaluated into a :class:`frozenset` and all elements must be
    keys of the `options` mapping argument.

    .. seealso::

       :class:`~.fields.BoundMultiValueField`
          is the :ref:`bound field class <api-aioxmpp.forms-bound-fields>` used
          by fields of this type.

       :class:`~.fields.AbstractChoiceField`
          for documentation on the `options` argument.

       :class:`~.fields.AbstractField`
          for documentation on the `var`, `type_`, `required`, `desc` and
          `label` arguments.

    """

    FIELD_TYPE = forms_xso.FieldType.LIST_MULTI

    def __init__(self, var, *, default=frozenset(), **kwargs):
        super().__init__(var, **kwargs)
        self._default = frozenset(default)
        if any(value not in self.options for value in self._default):
            raise ValueError(
                "invalid default: not in options"
            )

    def create_bound(self, for_instance):
        return BoundMultiSelectField(self, for_instance)

    def default(self):
        return self._default



########################################################################
# File name: test_xml.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import base64
import io
import itertools
import unittest
import random

import aioxmpp.xso as xso
import aioxmpp.xml

from aioxmpp.benchtest import times, timed, record


class ShallowRoot(xso.XSO):
    TAG = ("uri:test", "shallow")

    attr = xso.Attr("a")
    data = xso.Text()

    def __init__(self, scale=1):
        super().__init__()
        self.attr = "foobar"*(2*scale)
        self.data = "fnord"*(10*scale)


class DeepLeaf(xso.XSO):
    TAG = ("uri:test", "leaf")

    data = xso.Text()

    def generate(self, rng, depth):
        self.data = "foo" * (2*rng.randint(1, 10))


class DeepNode(xso.XSO):
    TAG = ("uri:test", "node")

    data = xso.Attr("attr")
    children = xso.ChildList([DeepLeaf])

    def generate(self, rng, depth):
        self.data = "foo" * (2*rng.randint(1, 10))
        if depth >= 5:
            cls = DeepLeaf
        else:
            cls = DeepNode

        self.children.append(cls())
        for i in range(rng.randint(2, 10)):
            if rng.randint(1, 10) == 1:
                item = DeepNode()
            else:
                item = DeepLeaf()
            self.children.append(item)

        for item in self.children:
            item.generate(rng, depth+1)


DeepNode.register_child(DeepNode.children, DeepNode)


class DeepRoot(xso.XSO):
    TAG = ("uri:test", "root")

    children = xso.ChildList([DeepLeaf, DeepNode])

    def generate(self, rng):
        self.children[:] = [DeepNode() for i in range(3)]
        for child in self.children:
            child.generate(rng, 1)


class TestxmlValidateNameValue_str(unittest.TestCase):
    KEY = "aioxmpp.xml", "xmlValidateNameValue"

    def test_exhaustive(self):
        validate = aioxmpp.xml.xmlValidateNameValue_str

        r1 = range(0, 0xd800)
        r2 = range(0xe000, 0xf0000)

        range_iter = itertools.chain(
            # exclude surrogates
            r1, r2,
        )

        with timed() as timer:
            for cp in range_iter:
                validate(chr(cp))

        record(self.KEY + ("exhaustive",),
               timer.elapsed / (len(r1) + len(r2)),
               "s")

    def test_exhaustive_dualchar(self):
        validate = aioxmpp.xml.xmlValidateNameValue_str

        strs = ["x" + chr(cp) for cp in range(0, 0xd800)]

        with timed() as timer:
            for s in strs:
                validate(s)

        record(self.KEY + ("exhaustive_dualchar",),
               timer.elapsed / (len(strs)),
               "s")

    def test_random_strings(self):
        key = self.KEY + ("random",)

        validate = aioxmpp.xml.xmlValidateNameValue_str

        rng = random.Random(1)
        samples = []
        for i in range(1000):
            samples.append(base64.b64encode(
                random.getrandbits(120).to_bytes(120//8, 'little')
            ).decode("ascii").rstrip("="))

        for sample in samples:
            with timed() as timer:
                validate(sample)
            record(key, timer.elapsed, "s")


class Testwrite_single_xso(unittest.TestCase):
    KEY = "aioxmpp.xml", "write_single_xso"

    @classmethod
    def setUpClass(cls):
        rng = random.Random(1)
        cls.deep_samples = [
            DeepRoot()
            for i in range(10)
        ]
        for sample in cls.deep_samples:
            with timed(cls.KEY+("deep", "generate")):
                sample.generate(rng)

    def setUp(self):
        self.buf = io.BytesIO(bytearray(1024*1024))

    def _reset_buffer(self):
        self.buf.seek(0)

    @times(1000)
    def test_shallow_and_small(self):
        key = self.KEY + ("shallow+small",)
        item = ShallowRoot()
        self._reset_buffer()
        with timed() as t:
            aioxmpp.xml.write_single_xso(item, self.buf)
        record(key+("sz",), self.buf.tell(), "B")
        record(key+("rate",), self.buf.tell() / t.elapsed, "B/s")

    @times(1000)
    def test_shallow_and_large(self):
        key = self.KEY + ("shallow+large",)
        item = ShallowRoot(scale=100)
        self._reset_buffer()
        with timed() as t:
            aioxmpp.xml.write_single_xso(item, self.buf)
        record(key+("sz",), self.buf.tell(), "B")
        record(key+("rate",), self.buf.tell() / t.elapsed, "B/s")

    @times(1000, pass_iteration=True)
    def test_deep(self, iteration=None):
        key = self.KEY + ("deep",)
        item = self.deep_samples[iteration % len(self.deep_samples)]
        self._reset_buffer()
        with timed() as t:
            aioxmpp.xml.write_single_xso(item, self.buf)
        record(key+("sz",), self.buf.tell(), "B")
        record(key+("rate",), self.buf.tell() / t.elapsed, "B/s")



########################################################################
# File name: main.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import sys

import PyQt5.Qt as Qt

import aioxmpp

from framework import Example

from .utils import asyncify_blocking
try:
    from .ui.main import Ui_MainWindow
except ImportError:
    print("You didn’t run make, I’ll try to do it for you...",
          file=sys.stderr)
    import subprocess
    try:
        subprocess.check_call(["make"])
        from .ui.main import Ui_MainWindow
    except Exception:
        print("Nope, that didn’t work out. "
              "You’ll have to fix that yourself. Sorry.",
              file=sys.stderr)
        raise

from .execute import Executor

if not hasattr(asyncio, "async"):
    setattr(asyncio, "async", asyncio.ensure_future)


class DiscoItemsModel(Qt.QAbstractTableModel):
    COLUMN_NAME = 0
    COLUMN_JID = 1
    COLUMN_NODE = 2
    COLUMN_COUNT = 3

    def __init__(self, parent=None):
        super().__init__(parent)
        self._items = []

    def replace(self, items):
        self.beginResetModel()
        self._items[:] = items
        self.endResetModel()

    def rowCount(self, parent):
        if parent.isValid():
            return 0
        return len(self._items)

    def columnCount(self, parent):
        return self.COLUMN_COUNT

    def data(self, index, role):
        if role != Qt.Qt.DisplayRole:
            return
        if not index.isValid():
            return

        item = self._items[index.row()]
        return {
            self.COLUMN_NAME: item.name,
            self.COLUMN_JID: str(item.jid),
            self.COLUMN_NODE: item.node or "",
        }.get(index.column())

    def headerData(self, section, orientation, role):
        if orientation != Qt.Qt.Horizontal:
            return
        if role != Qt.Qt.DisplayRole:
            return
        return {
            self.COLUMN_NAME: "Name",
            self.COLUMN_JID: "JID",
            self.COLUMN_NODE: "Node",
        }.get(section)


class MainWindow(Qt.QMainWindow):
    def __init__(self, close_fut):
        super().__init__()
        self.ui = Ui_MainWindow()
        self.ui.setupUi(self)
        self.ui.btn_scan.clicked.connect(
            self.scan
        )

        self.disco_model = DiscoItemsModel()
        self.ui.disco_items.setModel(self.disco_model)
        self.ui.disco_items.activated.connect(self.disco_item_activated)

        self.commands_model = DiscoItemsModel()
        self.ui.commands.setModel(self.commands_model)
        self.ui.commands.activated.connect(self.command_activated)

        self._close_fut = close_fut

    def disco_item_activated(self, index):
        if not index.isValid():
            return

        jid = self.disco_model.data(
            self.disco_model.index(
                index.row(),
                self.disco_model.COLUMN_JID,
                index.parent(),
            ),
            Qt.Qt.DisplayRole)

        self.ui.target_jid.setText(jid)
        self.scan()

    @asyncify_blocking
    @asyncio.coroutine
    def command_activated(self, index):
        if not index.isValid():
            return

        jid = aioxmpp.JID.fromstr(
            self.commands_model.data(
                self.commands_model.index(
                    index.row(),
                    self.disco_model.COLUMN_JID,
                    index.parent(),
                ),
                Qt.Qt.DisplayRole)
        )

        node = self.commands_model.data(
            self.commands_model.index(
                index.row(),
                self.disco_model.COLUMN_NODE,
                index.parent(),
            ),
            Qt.Qt.DisplayRole)

        session_fut = asyncio.ensure_future(
            self.adhoc_svc.execute(jid, node)
        )
        dlg = Executor(self)
        dlg.run_with_session_fut(node, session_fut)

    @asyncify_blocking
    @asyncio.coroutine
    def scan(self, *args, **kwargs):
        jid = aioxmpp.JID.fromstr(self.ui.target_jid.text())
        items = yield from self.disco_svc.query_items(
            jid
        )
        commands = yield from self.adhoc_svc.get_commands(
            jid
        )

        self.disco_model.replace(items.items)
        self.commands_model.replace(commands)

    def closeEvent(self, ev):
        if self._close_fut.done():
            return super().closeEvent(ev)
        self._close_fut.set_result(None)


class AdHocBrowser(Example):
    def __init__(self):
        super().__init__()
        self._close = asyncio.Future()
        self._mainwindow = MainWindow(self._close)

    @asyncio.coroutine
    def _main(self):
        self._mainwindow.client = self.client
        self._mainwindow.disco_svc = self.disco_svc
        self._mainwindow.adhoc_svc = self.adhoc_svc
        self._mainwindow.show()
        try:
            yield from self._close
        except:
            self._mainwindow.close()

    def _established(self):
        self._mainwindow.statusBar().showMessage(
            "connected as {}".format(self.client.local_jid)
        )
        if not self._mainwindow.ui.target_jid.text():
            self._mainwindow.ui.target_jid.setText(
                str(self.client.local_jid.domain)
            )

    @asyncio.coroutine
    def run_example(self):
        def kill(*args):
            nonlocal task
            if task is not None:
                task.cancel()

        self.client = self.make_simple_client()
        self.disco_svc = self.client.summon(aioxmpp.DiscoClient)
        self.adhoc_svc = self.client.summon(aioxmpp.AdHocClient)
        self.client.on_failure.connect(kill)
        self.client.on_stream_established.connect(self._established)

        cm = self.client.connected()
        aexit = cm.__aexit__
        yield from cm.__aenter__()
        try:
            task = asyncio.ensure_future(self._main())
            yield from task
        except:
            if not (yield from aexit(*sys.exc_info())):
                raise
        else:
            yield from aexit(None, None, None)



########################################################################
# File name: listen_pep.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import functools
import itertools
import io
import pathlib

import aioxmpp.disco
import aioxmpp.entitycaps
import aioxmpp.presence
import aioxmpp.xso

from framework import Example, exec_example


class ListenPEP(Example):
    def prepare_argparse(self):
        super().prepare_argparse()
        self.argparse.add_argument(
            "--namespace",
            dest="namespaces",
            default=[],
            action="append",
            metavar="PEP-NAMESPACE",
            help="PEP namespace to listen for (omit the +notify!). May be "
            "given multiple times."
        )

    def configure(self):
        super().configure()
        self.pep_namespaces = self.args.namespaces

    def _on_item_published(self, jid, node, item, message=None):
        buf = io.BytesIO()
        aioxmpp.xml.write_single_xso(item, buf)
        print(jid, node, buf.getvalue().decode("utf-8"))

    def make_simple_client(self):
        client = super().make_simple_client()
        self.caps = client.summon(aioxmpp.EntityCapsService)
        self.pep = client.summon(aioxmpp.PEPClient)

        self.claims = []
        for ns in self.pep_namespaces:
            claim = self.pep.claim_pep_node(
                ns,
                notify=True,
            )
            claim.on_item_publish.connect(self._on_item_published)
            self.claims.append(claim)

        return client

    @asyncio.coroutine
    def run_example(self):
        self.stop_event = self.make_sigint_event()
        yield from super().run_example()

    @asyncio.coroutine
    def run_simple_example(self):
        yield from self.stop_event.wait()


if __name__ == "__main__":
    exec_example(ListenPEP())



########################################################################
# File name: xml.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
"""
:mod:`~aioxmpp.xml` --- XML utilities and interfaces for handling XMPP XML streams
#######################################################################################

This module provides a few classes and functions which are useful when
generating and parsing XML streams for XMPP.

Generating XML streams
======================

The most useful class here is the :class:`XMPPXMLGenerator`:

.. autoclass:: XMPPXMLGenerator

.. autoclass:: XMLStreamWriter

Processing XML streams
======================

To convert streams of SAX events to :class:`~.stanza_model.XSO`
instances, the following classes and functions can be used:

.. autoclass:: XMPPXMLProcessor

.. autoclass:: XMPPLexicalHandler

.. autofunction:: make_parser

Utility functions
=================

.. autofunction:: serialize_single_xso

.. autofunction:: write_single_xso

.. autofunction:: read_xso

.. autofunction:: read_single_xso

"""

import copy
import ctypes
import ctypes.util
import contextlib
import io

import xml.sax
import xml.sax.saxutils

from enum import Enum

from . import errors, structs, xso
from .utils import namespaces


_NAME_START_CHAR = [
    [ord(":"), ord("_")],
    range(ord("a"), ord("z")+1),
    range(ord("A"), ord("Z")+1),
    range(0xc0, 0xd7),
    range(0xd8, 0xf7),
    range(0xf8, 0x300),
    range(0x370, 0x37e),
    range(0x37f, 0x2000),
    range(0x200c, 0x200e),
    range(0x2070, 0x2190),
    range(0x2c00, 0x2ff0),
    range(0x3001, 0xd800),
    range(0xf900, 0xfdd0),
    range(0xfdf0, 0xfffe),
    range(0x10000, 0xf0000),
]

_NAME_CHAR = _NAME_START_CHAR + [
    [ord("-"), ord("."), 0xb7],
    range(ord("0"), ord("9")+1),
    range(0x0300, 0x0370),
    range(0x203f, 0x2041),
]
_NAME_CHAR.sort(key=lambda x: x[0])


def xmlValidateNameValue_str(s):
    if not s:
        return False
    ch = ord(s[0])
    if not any(ch in range_ for range_ in _NAME_START_CHAR):
        return False
    return all(
        any(ch in range_ for range_ in _NAME_CHAR)
        for ch in map(ord, s)
    )


def is_valid_cdata_str(s):
    for c in s:
        o = ord(c)
        if o >= 32:
            continue
        if o < 9 or 11 <= o <= 12 or 14 <= o <= 31:
            return False

    return True


class XMPPXMLGenerator:
    """
    Class to generate XMPP-conforming XML bytes.

    :param out: File-like object to which the bytes are written.
    :param short_empty_elements: Write empty elements as ``<foo/>`` instead of
        ``<foo></foo>``.
    :type short_empty_elements: :class:`bool`
    :param sorted_attributes: Sort the attributes in the output. Note: this
        comes with a performance penalty. See below.
    :type sorted_attributes: :class:`bool`
    :param additional_escapes: Sequence of characters to escape in CDATA.
    :type additional_escapes: :class:`~collections.abc.Iterable` of
        1-codepoint :class:`str` objects.

    :class:`XMPPXMLGenerator` works similar to
    :class:`xml.sax.saxutils.XMLGenerator`, but has a few key differences:

    * It supports **only** namespace-conforming XML documents
    * It automatically chooses namespace prefixes if a namespace has not been
      declared, while avoiding to use prefixes at all if possible
    * It is in general stricter on (explicit) namespace declarations, to avoid
      ambiguities
    * It always uses utf-8 ☺
    * It allows explicit flushing

    `out` must be a file-like supporting both :meth:`file.write` and
    :meth:`file.flush`.

    If `short_empty_elements` is true, empty elements are rendered as
    ``<foo/>`` instead of ``<foo></foo>``, unless a flush occurs before the
    call to :meth:`endElementNS`, in which case the opening is finished before
    flushing, thus the long form is generated.

    If `sorted_attributes` is true, attributes are emitted in the lexical order
    of their qualified names (except for namespace declarations, which are
    always sorted and always before the normal attributes). The default is not
    to do this, for performance. During testing, however, it is useful to have
    a consistent oder on the attributes.

    All characters in `additional_escapes` are escaped using XML entities. Note
    that ``<``, ``>`` and ``&`` are always escaped. `additional_escapes` is
    converted to a dictionary for use with :func:`~xml.sax.saxutils.escape` and
    :func:`~xml.sax.saxutils.quoteattr`. Passing a dictionary to
    `additional_escapes` or passing multi-character strings as elements of
    `additional_escapes` is **not** supported since it may be (ab-)used to
    create invalid XMPP XML. `additional_escapes` affects both CDATA in XML
    elements as well as attribute values.

    Implementation of the SAX content handler interface (see
    :class:`xml.sax.handler.ContentHandler`):

    .. automethod:: startDocument

    .. automethod:: startPrefixMapping(prefix, uri)

    .. automethod:: startElementNS

    .. automethod:: characters

    .. automethod:: endElementNS

    .. automethod:: endPrefixMapping

    .. automethod:: endDocument

    The following SAX content handler methods have deliberately not been
    implemented:

    .. automethod:: setDocumentLocator

    .. automethod:: skippedEntity

    .. automethod:: ignorableWhitespace

    .. automethod:: startElement

    .. automethod:: endElement

    These methods produce content which is invalid in XMPP XML streams and thus
    always raise :class:`ValueError`:

    .. automethod:: processingInstruction

    In addition to the SAX content handler interface, the following methods are
    provided:

    .. automethod:: flush

    .. automethod:: buffer

    """
    def __init__(self, out,
                 short_empty_elements=True,
                 sorted_attributes=False,
                 additional_escapes=[]):
        self._write = out.write
        if hasattr(out, "flush"):
            self._flush = out.flush
        else:
            self._flush = None

        self._short_empty_elements = short_empty_elements
        self._sorted_attributes = sorted_attributes

        self._additional_escapes = {
            char: "&#{};".format(ord(char))
            for char in additional_escapes
        }

        # NOTE: when adding state, make sure to handle it in buffer() and to
        # add tests that buffer() handles it correctly
        self._ns_map_stack = [({}, set(), 0)]
        self._curr_ns_map = {}
        self._pending_start_element = False
        self._ns_prefixes_floating_in = {}
        self._ns_prefixes_floating_out = set()
        self._ns_auto_prefixes_floating_in = set()
        self._ns_decls_floating_in = {}
        self._ns_counter = -1

        # for buffer()
        self._buf = None
        self._buf_in_use = False

    def _roll_prefix(self, attr):
        if not attr and None not in self._ns_prefixes_floating_in:
            return None

        prefix_number = self._ns_counter + 1
        while True:
            prefix = "ns{}".format(prefix_number)
            if prefix not in self._ns_prefixes_floating_in:
                break
            prefix_number += 1
        self._ns_counter = prefix_number
        return prefix

    def _qname(self, name, attr=False):
        if not isinstance(name, tuple):
            raise ValueError("names must be tuples")

        if ":" in name[1] or not xmlValidateNameValue_str(name[1]):
            raise ValueError("invalid name: {!r}".format(name[1]))

        if name[0]:
            if name[0] == "http://www.w3.org/XML/1998/namespace":
                return "xml:" + name[1]
            try:
                prefix = self._ns_decls_floating_in[name[0]]
                if attr and prefix is None:
                    raise KeyError()
            except KeyError:
                try:
                    prefix = self._curr_ns_map[name[0]]
                    if prefix in self._ns_prefixes_floating_in:
                        raise KeyError()
                    if attr and prefix is None:
                        raise KeyError()
                except KeyError:
                    # namespace is undeclared, we have to declare it..
                    prefix = self._roll_prefix(attr)
                    self.startPrefixMapping(prefix, name[0], auto=True)

            if prefix:
                return ":".join((prefix, name[1]))

        elif   (not attr and
                (None in self._curr_ns_map or
                 None in self._ns_prefixes_floating_in)):
            raise ValueError("cannot create unnamespaced element when "
                             "prefixless namespace is bound")

        return name[1]

    def _finish_pending_start_element(self):
        if not self._pending_start_element:
            return
        self._pending_start_element = False
        self._write(b">")

    def _pin_floating_ns_decls(self, old_counter):
        if self._ns_prefixes_floating_out:
            raise RuntimeError("namespace prefix has not been closed")

        new_decls = self._ns_decls_floating_in
        new_prefixes = self._ns_prefixes_floating_in
        self._ns_map_stack.append(
            (
                self._curr_ns_map.copy(),
                set(new_prefixes) - self._ns_auto_prefixes_floating_in,
                old_counter
            )
        )

        cleared_new_prefixes = dict(new_prefixes)
        for uri, prefix in self._curr_ns_map.items():
            try:
                new_uri = cleared_new_prefixes[prefix]
            except KeyError:
                pass
            else:
                if new_uri == uri:
                    del cleared_new_prefixes[prefix]

        self._curr_ns_map.update(new_decls)
        self._ns_decls_floating_in = {}
        self._ns_prefixes_floating_in = {}
        self._ns_auto_prefixes_floating_in.clear()

        return cleared_new_prefixes

    def startDocument(self):
        """
        Start the document. This method *must* be called before any other
        content handler method.
        """
        # yes, I know the doctext is not enforced. It might become enforced in
        # a later version though, when I find a compelling reason why it is
        # needed.
        self._write(b'<?xml version="1.0"?>')

    def startPrefixMapping(self, prefix, uri, *, auto=False):
        """
        Start a prefix mapping which maps the given `prefix` to the given
        `uri`.

        Note that prefix mappings are handled transactional. All announcements
        of prefix mappings are collected until the next call to
        :meth:`startElementNS`. At that point, the mappings are collected and
        start to override the previously declared mappings until the
        corresponding :meth:`endElementNS` call.

        Also note that calling :meth:`startPrefixMapping` is not mandatory; you
        can use any namespace you like at any time. If you use a namespace
        whose URI has not been associated with a prefix yet, a free prefix will
        automatically be chosen. To avoid unneccessary performance penalties,
        do not use prefixes of the form ``"ns{:d}".format(n)``, for any
        non-negative number of `n`.

        It is however required to call :meth:`endPrefixMapping` after a
        :meth:`endElementNS` call for all namespaces which have been announced
        directly before the :meth:`startElementNS` call (except for those which
        have been chosen automatically). Not doing so will result in a
        :class:`RuntimeError` at the next :meth:`startElementNS` or
        :meth:`endElementNS` call.

        During a transaction, it is not allowed to declare the same prefix
        multiple times.
        """
        if     (prefix is not None and
                (prefix == "xml" or
                 prefix == "xmlns" or
                 not xmlValidateNameValue_str(prefix) or
                 ":" in prefix)):
            raise ValueError("not a valid prefix: {!r}".format(prefix))

        if prefix in self._ns_prefixes_floating_in:
            raise ValueError("prefix already declared for next element")
        if auto:
            self._ns_auto_prefixes_floating_in.add(prefix)
        self._ns_prefixes_floating_in[prefix] = uri
        self._ns_decls_floating_in[uri] = prefix

    def startElementNS(self, name, qname, attributes=None):
        """
        Start a sub-element. `name` must be a tuple of ``(namespace_uri,
        localname)`` and `qname` is ignored. `attributes` must be a dictionary
        mapping attribute tag tuples (``(namespace_uri, attribute_name)``) to
        string values. To use unnamespaced attributes, `namespace_uri` can be
        false (e.g. :data:`None` or the empty string).

        To use unnamespaced elements, `namespace_uri` in `name` must be false
        **and** no namespace without prefix must be currently active. If a
        namespace without prefix is active and `namespace_uri` in `name` is
        false, :class:`ValueError` is raised.

        Attribute values are of course automatically escaped.
        """
        self._finish_pending_start_element()
        old_counter = self._ns_counter

        qname = self._qname(name)
        if attributes:
            attrib = [
                (self._qname(attrname, attr=True), value)
                for attrname, value in attributes.items()
            ]
            for attrqname, _ in attrib:
                if attrqname == "xmlns":
                    raise ValueError("xmlns not allowed as attribute name")
        else:
            attrib = []

        pending_prefixes = self._pin_floating_ns_decls(old_counter)

        self._write(b"<")
        self._write(qname.encode("utf-8"))

        if None in pending_prefixes:
            uri = pending_prefixes.pop(None)
            self._write(b" xmlns=")
            self._write(xml.sax.saxutils.quoteattr(uri).encode("utf-8"))

        for prefix, uri in sorted(pending_prefixes.items()):
            self._write(b" xmlns")
            if prefix:
                self._write(b":")
                self._write(prefix.encode("utf-8"))
            self._write(b"=")
            self._write(
                xml.sax.saxutils.quoteattr(uri).encode("utf-8")
            )

        if self._sorted_attributes:
            attrib.sort()

        for attrname, value in attrib:
            self._write(b" ")
            self._write(attrname.encode("utf-8"))
            self._write(b"=")
            self._write(
                xml.sax.saxutils.quoteattr(
                    value,
                    self._additional_escapes,
                ).encode("utf-8")
            )

        if self._short_empty_elements:
            self._pending_start_element = name
        else:
            self._write(b">")

    def endElementNS(self, name, qname):
        """
        End a previously started element. `name` must be a ``(namespace_uri,
        localname)`` tuple and `qname` is ignored.
        """
        if self._ns_prefixes_floating_out:
            raise RuntimeError("namespace prefix has not been closed")

        if self._pending_start_element == name:
            self._pending_start_element = False
            self._write(b"/>")
        else:
            self._write(b"</")
            self._write(self._qname(name).encode("utf-8"))
            self._write(b">")

        self._curr_ns_map, self._ns_prefixes_floating_out, self._ns_counter = \
            self._ns_map_stack.pop()

    def endPrefixMapping(self, prefix):
        """
        End a prefix mapping declared with :meth:`startPrefixMapping`. See
        there for more details.
        """
        self._ns_prefixes_floating_out.remove(prefix)

    def startElement(self, name, attributes=None):
        """
        Not supported; only elements with proper namespacing are supported by
        this generator.
        """
        raise NotImplementedError("namespace-incorrect documents are "
                                  "not supported")

    def characters(self, chars):
        """
        Put character data in the currently open element. Special characters
        (such as ``<``, ``>`` and ``&``) are escaped.

        If `chars` contains any ASCII control character, :class:`ValueError` is
        raised.
        """
        self._finish_pending_start_element()
        if not is_valid_cdata_str(chars):
            raise ValueError("control characters are not allowed in "
                             "well-formed XML")
        self._write(xml.sax.saxutils.escape(
            chars,
            self._additional_escapes,
        ).encode("utf-8"))

    def processingInstruction(self, target, data):
        """
        Not supported; explicitly forbidden in XMPP. Raises
        :class:`ValueError`.
        """
        raise ValueError("restricted xml: processing instruction forbidden")

    def skippedEntity(self, name):
        """
        Not supported; there is no use case. Raises
        :class:`NotImplementedError`.
        """
        raise NotImplementedError("skippedEntity")

    def setDocumentLocator(self, locator):
        """
        Not supported; there is no use case. Raises
        :class:`NotImplementedError`.
        """
        raise NotImplementedError("setDocumentLocator")

    def ignorableWhitespace(self, whitespace):
        """
        Not supported; could be mapped to :meth:`characters`.
        """
        raise NotImplementedError("ignorableWhitespace")

    def endElement(self, name):
        """
        Not supported; only elements with proper namespacing are supported by
        this generator.
        """
        self.startElement(name)

    def endDocument(self):
        """
        This must be called at the end of the document. Note that this does not
        call :meth:`flush`.
        """

    def flush(self):
        """
        Call :meth:`flush` on the object passed to the `out` argument of the
        constructor. In addition, any unfinished opening tags are finished,
        which can lead to expansion of the generated XML code (see note on the
        `short_empty_elements` argument at the class documentation).
        """
        self._finish_pending_start_element()
        if self._flush:
            self._flush()

    @contextlib.contextmanager
    def _save_state(self):
        """
        Helper context manager for :meth:`buffer` which saves the whole state.

        This is broken out in a separate method for readability and tested
        indirectly by testing :meth:`buffer`.
        """
        ns_prefixes_floating_in = copy.copy(self._ns_prefixes_floating_in)
        ns_prefixes_floating_out = copy.copy(self._ns_prefixes_floating_out)
        ns_decls_floating_in = copy.copy(self._ns_decls_floating_in)
        curr_ns_map = copy.copy(self._curr_ns_map)
        ns_map_stack = copy.copy(self._ns_map_stack)
        pending_start_element = self._pending_start_element
        ns_counter = self._ns_counter
        # XXX: I have been unable to find a test justifying copying this :/
        # for completeness, I’m still doing it
        ns_auto_prefixes_floating_in = \
            copy.copy(self._ns_auto_prefixes_floating_in)
        try:
            yield
        except:
            self._ns_prefixes_floating_in = ns_prefixes_floating_in
            self._ns_prefixes_floating_out = ns_prefixes_floating_out
            self._ns_decls_floating_in = ns_decls_floating_in
            self._pending_start_element = pending_start_element
            self._curr_ns_map = curr_ns_map
            self._ns_map_stack = ns_map_stack
            self._ns_counter = ns_counter
            self._ns_auto_prefixes_floating_in = ns_auto_prefixes_floating_in
            raise

    @contextlib.contextmanager
    def buffer(self):
        """
        Context manager to temporarily buffer the output.

        :raise RuntimeError: If two :meth:`buffer` context managers are used
                             nestedly.

        If the context manager is left without exception, the buffered output
        is sent to the actual sink. Otherwise, it is discarded.

        In addition to the output being buffered, buffer also captures the
        entire state of the XML generator and restores it to the previous state
        if the context manager is left with an exception.

        This can be used to fail-safely attempt to serialise a subtree and
        return to a well-defined state if serialisation fails.

        :meth:`flush` is not called automatically.

        If :meth:`flush` is called while a :meth:`buffer` context manager is
        active, no actual flushing happens (but unfinished opening tags are
        closed as usual, see the `short_empty_arguments` parameter).
        """
        if self._buf_in_use:
            raise RuntimeError("nested use of buffer() is not supported")
        self._buf_in_use = True
        old_write = self._write
        old_flush = self._flush

        if self._buf is None:
            self._buf = io.BytesIO()
        else:
            try:
                self._buf.seek(0)
                self._buf.truncate()
            except BufferError:
                # we need a fresh buffer for this, the other is still in use.
                self._buf = io.BytesIO()

        self._write = self._buf.write
        self._flush = None
        try:
            with self._save_state():
                yield
            old_write(self._buf.getbuffer())
            if old_flush:
                old_flush()
        finally:
            self._buf_in_use = False
            self._write = old_write
            self._flush = old_flush


class XMLStreamWriter:
    """
    A convenient class to write a standard conforming XML stream.

    :param f: File-like object to write to.
    :param to: Address to which the connection is addressed.
    :type to: :class:`aioxmpp.JID`
    :param from_: Optional address from which the connection originates.
    :type from_: :class:`aioxmpp.JID`
    :param version: Version of the XML stream protocol.
    :type version: :class:`tuple` of (:class:`int`, :class:`int`)
    :param nsmap: Mapping of namespaces to declare at the stream header.

    .. note::

       The constructor *does not* send a stream header. :meth:`start` must be
       called explicitly to send a stream header.

    The generated stream header follows :rfc:`6120` and has the ``to`` and
    ``version`` attributes as well as optionally the ``from`` attribute
    (controlled by `from_`). In addition, the namespace prefixes defined by
    `nsmap` (mapping prefixes to namespace URIs) are declared on the stream
    header.

    .. note::

       It is unfortunately not allowed to use namespace prefixes in stanzas
       which were declared in stream headers as convenient as that would be.
       The option is thus only useful to declare the default namespace for
       stanzas.

    .. autoattribute:: closed

    The following methods are used to generate output:

    .. automethod:: start

    .. automethod:: send

    .. automethod:: abort

    .. automethod:: close
    """

    def __init__(self, f, to,
                 from_=None,
                 version=(1, 0),
                 nsmap={},
                 sorted_attributes=False):
        super().__init__()
        self._to = to
        self._from = from_
        self._version = version
        self._writer = XMPPXMLGenerator(
            out=f,
            short_empty_elements=True,
            sorted_attributes=sorted_attributes)
        self._nsmap_to_use = {
            "stream": namespaces.xmlstream
        }
        self._nsmap_to_use.update(nsmap)
        self._closed = False

    @property
    def closed(self):
        """
        True if the stream has been closed by :meth:`abort` or :meth:`close`.
        Read-only.
        """
        return self._closed

    def start(self):
        """
        Send the stream header as described above.
        """
        attrs = {
            (None, "to"): str(self._to),
            (None, "version"): ".".join(map(str, self._version))
        }
        if self._from:
            attrs[None, "from"] = str(self._from)

        self._writer.startDocument()
        for prefix, uri in self._nsmap_to_use.items():
            self._writer.startPrefixMapping(prefix, uri)
        self._writer.startElementNS(
            (namespaces.xmlstream, "stream"),
            None,
            attrs)
        self._writer.flush()

    def send(self, xso):
        """
        Send a single XML stream object.

        :param xso: Object to serialise and send.
        :type xso: :class:`aioxmpp.xso.XSO`
        :raises Exception: from any serialisation errors, usually
                           :class:`ValueError`.

        Serialise the `xso` and send it over the stream. If any serialisation
        error occurs, no data is sent over the stream and the exception is
        re-raised; the :meth:`send` method thus provides strong exception
        safety.

        .. warning::

           The behaviour of :meth:`send` after :meth:`abort` or :meth:`close`
           and before :meth:`start` is undefined.

        """
        with self._writer.buffer():
            xso.unparse_to_sax(self._writer)

    def abort(self):
        """
        Abort the stream.

        The stream is flushed and the internal data structures are cleaned up.
        No stream footer is sent. The stream is :attr:`closed` afterwards.

        If the stream is already :attr:`closed`, this method does nothing.
        """
        if self._closed:
            return
        self._closed = True
        self._writer.flush()
        del self._writer

    def close(self):
        """
        Close the stream.

        The stream footer is sent and the internal structures are cleaned up.

        If the stream is already :attr:`closed`, this method does nothing.
        """
        if self._closed:
            return
        self._closed = True
        self._writer.endElementNS((namespaces.xmlstream, "stream"), None)
        for prefix in self._nsmap_to_use:
            self._writer.endPrefixMapping(prefix)
        self._writer.endDocument()
        del self._writer


class ProcessorState(Enum):
    CLEAN = 0
    STARTED = 1
    STREAM_HEADER_PROCESSED = 2
    STREAM_FOOTER_PROCESSED = 3
    EXCEPTION_BACKOFF = 4


class XMPPXMLProcessor:
    """
    This class is a :class:`xml.sax.handler.ContentHandler`. It
    can be used to parse an XMPP XML stream.

    When used with a :class:`xml.sax.xmlreader.XMLReader`, it gradually
    processes the incoming XML stream. If any restricted XML is encountered, an
    appropriate :class:`~.errors.StreamError` is raised.

    .. warning::

       To achieve compliance with XMPP, it is recommended to use
       :class:`XMPPLexicalHandler` as lexical handler, using
       :meth:`xml.sax.xmlreader.XMLReader.setProperty`::

            parser.setProperty(xml.sax.handler.property_lexical_handler,
                               XMPPLexicalHandler)

       Otherwise, invalid XMPP XML such as comments, entity references and DTD
       declarations will not be caught.

    **Exception handling**: When an exception occurs while parsing a
    stream-level element, such as a stanza, the exception is stored internally
    and exception handling is invoked. During exception handling, all SAX
    events are dropped, until the stream-level element has been completely
    processed by the parser. Then, if available, :attr:`on_exception` is
    called, with the stored exception as the only argument. If
    :attr:`on_exception` is false (e.g. :data:`None`), the exception is
    re-raised from the :meth:`endElementNS` handler, in turn most likely
    destroying the SAX parsers internal state.

    .. attribute:: on_exception

       May be a callable or :data:`None`. If not false, the value will get
       called when exception handling has finished, with the exception as the
       only argument.

    .. attribute:: on_stream_footer

       May be a callable or :data:`None`. If not false, the value will get
       called whenever a stream footer is processed.

    .. attribute:: on_stream_header

       May be a callable or :data:`None`. If not false, the value will get
       called whenever a stream header is processed.

    .. autoattribute:: stanza_parser
    """

    def __init__(self):
        super().__init__()
        self._state = ProcessorState.CLEAN
        self._stanza_parser = None
        self._stored_exception = None
        self.on_stream_header = None
        self.on_stream_footer = None
        self.on_exception = None

        self.remote_version = None
        self.remote_from = None
        self.remote_to = None
        self.remote_id = None
        self.remote_lang = None

    @property
    def stanza_parser(self):
        """
        A :class:`~.xso.XSOParser` object (or compatible) which will
        receive the sax-ish events used in :mod:`~aioxmpp.xso`. It
        is driven using an instance of :class:`~.xso.SAXDriver`.

        This object can only be set before :meth:`startDocument` has been
        called (or after :meth:`endDocument` has been called).
        """
        return self._stanza_parser

    @stanza_parser.setter
    def stanza_parser(self, value):
        if self._state != ProcessorState.CLEAN:
            raise RuntimeError("invalid state: {}".format(self._state))
        self._stanza_parser = value
        self._stanza_parser.lang = self.remote_lang

    def processingInstruction(self, target, foo):
        raise errors.StreamError(
            errors.StreamErrorCondition.RESTRICTED_XML,
            "processing instructions are not allowed in XMPP"
        )

    def characters(self, characters):
        if self._state == ProcessorState.EXCEPTION_BACKOFF:
            pass
        elif self._state != ProcessorState.STREAM_HEADER_PROCESSED:
            raise RuntimeError("invalid state: {}".format(self._state))
        else:
            self._driver.characters(characters)

    def startDocument(self):
        if self._state != ProcessorState.CLEAN:
            raise RuntimeError("invalid state: {}".format(self._state))
        self._state = ProcessorState.STARTED
        self._depth = 0
        self._driver = xso.SAXDriver(self._stanza_parser)

    def startElement(self, name, attributes):
        raise RuntimeError("incorrectly configured parser: "
                           "startElement called (instead of startElementNS)")

    def endElement(self, name):
        raise RuntimeError("incorrectly configured parser: "
                           "endElement called (instead of endElementNS)")

    def endDocument(self):
        if self._state != ProcessorState.STREAM_FOOTER_PROCESSED:
            raise RuntimeError("invalid state: {}".format(self._state))
        self._state = ProcessorState.CLEAN
        self._driver = None

    def startPrefixMapping(self, prefix, uri):
        pass

    def endPrefixMapping(self, prefix):
        pass

    def startElementNS(self, name, qname, attributes):
        if self._state == ProcessorState.STREAM_HEADER_PROCESSED:
            try:
                self._driver.startElementNS(name, qname, attributes)
            except Exception as exc:
                self._stored_exception = exc
                self._state = ProcessorState.EXCEPTION_BACKOFF
            self._depth += 1
            return
        elif self._state == ProcessorState.EXCEPTION_BACKOFF:
            self._depth += 1
            return
        elif self._state != ProcessorState.STARTED:
            raise RuntimeError("invalid state: {}".format(self._state))

        if name != (namespaces.xmlstream, "stream"):
            raise errors.StreamError(
                errors.StreamErrorCondition.INVALID_NAMESPACE,
                "stream has invalid namespace or localname"
            )

        attributes = dict(attributes)
        try:
            self.remote_version = tuple(
                map(int, attributes.pop((None, "version"), "0.9").split("."))
            )
        except ValueError as exc:
            raise errors.StreamError(
                errors.StreamErrorCondition.UNSUPPORTED_VERSION,
                str(exc)
            )

        remote_to = attributes.pop((None, "to"), None)
        if remote_to is not None:
            remote_to = structs.JID.fromstr(remote_to)
        self.remote_to = remote_to

        try:
            self.remote_from = structs.JID.fromstr(
                attributes.pop((None, "from"))
            )
        except KeyError:
            raise errors.StreamError(
                errors.StreamErrorCondition.UNDEFINED_CONDITION,
                "from attribute required in response header"
            )
        try:
            self.remote_id = attributes.pop((None, "id"))
        except KeyError:
            raise errors.StreamError(
                errors.StreamErrorCondition.UNDEFINED_CONDITION,
                "id attribute required in response header"
            )

        try:
            lang = attributes.pop((namespaces.xml, "lang"))
        except KeyError:
            self.remote_lang = None
        else:
            self.remote_lang = structs.LanguageTag.fromstr(lang)

        if self._stanza_parser is not None:
            self._stanza_parser.lang = self.remote_lang

        if self.on_stream_header:
            self.on_stream_header()

        self._state = ProcessorState.STREAM_HEADER_PROCESSED
        self._depth += 1

    def _end_element_exception_handling(self):
        self._state = ProcessorState.STREAM_HEADER_PROCESSED
        exc = self._stored_exception
        self._stored_exception = None
        if self.on_exception:
            self.on_exception(exc)
        else:
            raise exc

    def endElementNS(self, name, qname):
        if self._state == ProcessorState.STREAM_HEADER_PROCESSED:
            self._depth -= 1
            if self._depth > 0:
                try:
                    return self._driver.endElementNS(name, qname)
                except Exception as exc:
                    self._stored_exception = exc
                    self._state = ProcessorState.EXCEPTION_BACKOFF
                    if self._depth == 1:
                        self._end_element_exception_handling()
            else:
                if self.on_stream_footer:
                    self.on_stream_footer()
                self._state = ProcessorState.STREAM_FOOTER_PROCESSED

        elif self._state == ProcessorState.EXCEPTION_BACKOFF:
            self._depth -= 1
            if self._depth == 1:
                self._end_element_exception_handling()
        else:
            raise RuntimeError("invalid state: {}".format(self._state))


class XMPPLexicalHandler:
    """
    A `lexical handler
    <http://www.saxproject.org/apidoc/org/xml/sax/ext/LexicalHandler.html>`_
    which rejects certain contents which are invalid in an XMPP XML stream:

    * comments,
    * dtd declarations,
    * non-predefined entities.

    The class can be used as lexical handler directly; all methods are
    stateless and can be used both on the class and on objects of the class.

    """
    PREDEFINED_ENTITIES = {"amp", "lt", "gt", "apos", "quot"}

    @classmethod
    def comment(cls, data):
        raise errors.StreamError(
            errors.StreamErrorCondition.RESTRICTED_XML,
            "comments are not allowed in XMPP"
        )

    @classmethod
    def startDTD(cls, name, publicId, systemId):
        raise errors.StreamError(
            errors.StreamErrorCondition.RESTRICTED_XML,
            "DTD declarations are not allowed in XMPP"
        )

    @classmethod
    def endDTD(cls):
        pass

    @classmethod
    def startCDATA(cls):
        pass

    @classmethod
    def endCDATA(cls):
        pass

    @classmethod
    def startEntity(cls, name):
        if name not in cls.PREDEFINED_ENTITIES:
            raise errors.StreamError(
                errors.StreamErrorCondition.RESTRICTED_XML,
                "non-predefined entities are not allowed in XMPP"
            )

    @classmethod
    def endEntity(cls, name):
        pass


def make_parser():
    """
    Create a parser which is suitably configured for parsing an XMPP XML
    stream. It comes equipped with :class:`XMPPLexicalHandler`.
    """
    p = xml.sax.make_parser()
    p.setFeature(xml.sax.handler.feature_namespaces, True)
    p.setFeature(xml.sax.handler.feature_external_ges, False)
    p.setProperty(xml.sax.handler.property_lexical_handler,
                  XMPPLexicalHandler)
    return p


def serialize_single_xso(x):
    """
    Serialize a single XSO `x` to a string. This is potentially very slow and
    should only be used for debugging purposes. It is generally more efficient
    to use a :class:`XMPPXMLGenerator` to stream elements.
    """
    buf = io.BytesIO()
    gen = XMPPXMLGenerator(buf,
                           short_empty_elements=True,
                           sorted_attributes=True)
    x.unparse_to_sax(gen)
    return buf.getvalue().decode("utf8")


def write_single_xso(x, dest):
    """
    Write a single XSO `x` to a binary file-like object `dest`.
    """
    gen = XMPPXMLGenerator(dest,
                           short_empty_elements=True,
                           sorted_attributes=True)
    x.unparse_to_sax(gen)


def read_xso(src, xsomap):
    """
    Read a single XSO from a binary file-like input `src` containing an XML
    document.

    `xsomap` must be a mapping which maps :class:`~.XSO` subclasses
    to callables. These will be registered at a newly created
    :class:`.xso.XSOParser` instance which will be used to parse the  document
    in `src`.

    The `xsomap` is thus used to determine the class parsing the root element
    of the XML document. This can be used to support multiple versions.
    """

    xso_parser = xso.XSOParser()

    for class_, cb in xsomap.items():
        xso_parser.add_class(class_, cb)

    driver = xso.SAXDriver(xso_parser)

    parser = xml.sax.make_parser()
    parser.setFeature(
        xml.sax.handler.feature_namespaces,
        True)
    parser.setFeature(
        xml.sax.handler.feature_external_ges,
        False)
    parser.setContentHandler(driver)

    parser.parse(src)


def read_single_xso(src, type_):
    """
    Read a single :class:`~.XSO` of the given `type_` from the binary file-like
    input `src` and return the instance.
    """

    result = None

    def cb(instance):
        nonlocal result
        result = instance

    read_xso(src, {type_: cb})

    return result
